from nbodykit.extensionpoints import Algorithm, DataSource
from nbodykit.fkp import FKPCatalog
import numpy
import logging

class BianchiPowerAlgorithm(Algorithm):
    """
    Algorithm to compute the power spectrum multipoles using FFTs
    for a data survey with non-trivial geometry
    
    The algorithm used to compute the multipoles is detailed
    in Bianchi et al. 2015 (http://adsabs.harvard.edu/abs/2015MNRAS.453L..11B)
    """
    plugin_name = "BianchiFFTPower"
    logger = logging.getLogger(plugin_name)

    def __init__(self, data, randoms, Nmesh, max_ell, 
                    paintbrush='cic',
                    dk=None, 
                    kmin=0., 
                    BoxSize=None, 
                    BoxPad=0.02, 
                    compute_fkp_weights=False, 
                    P0_fkp=None, 
                    nbar=None, 
                    fsky=None,
                    factor_hexadecapole=False,
                    keep_cache=False):
                           
        # initialize the FKP catalog (unopened)
        kws = {}
        kws['BoxSize'] = self.BoxSize
        kws['BoxPad'] = self.BoxPad
        kws['compute_fkp_weights'] = self.compute_fkp_weights
        kws['P0_fkp'] = self.P0_fkp
        kws['nbar'] = self.nbar
        kws['fsky'] = self.fsky
        self.catalog = FKPCatalog(self.data, self.randoms, **kws)
        
    @classmethod
    def register(cls):
        
        s = cls.schema
        s.description = "power spectrum multipoles using FFTs for a data survey with \n"
        s.description += "non-trivial geometry, as detailed in Bianchi et al. 2015"

        # the required arguments
        s.add_argument("data", type=DataSource.from_config,
            help="DataSource representing the `data` catalog")
        s.add_argument("randoms", type=DataSource.from_config,
            help="DataSource representing the `randoms` catalog")        
        s.add_argument("Nmesh", type=int,
            help='the number of cells in the gridded mesh (per axis)')
        s.add_argument('max_ell', type=int, choices=[0,2,4],
            help='compute multipoles up to and including this ell value')

        # the optional arguments
        s.add_argument("BoxSize", type=DataSource.BoxSizeParser,
            help="the size of the box; if not provided, automatically computed from the `randoms` catalog")
        s.add_argument('BoxPad', type=float,
            help='when setting the box size automatically, apply this additional buffer')
        s.add_argument('compute_fkp_weights', type=bool,
            help='if set, use FKP weights, computed from `P0_fkp` and the provided `nbar`') 
        s.add_argument('P0_fkp', type=float,
            help='the fiducial power value `P0` used to compute FKP weights')
        s.add_argument('nbar', type=str,
            help='read `nbar(z)` from this file, which provides two columns (z, nbar)')
        s.add_argument('fsky', type=float, 
            help='the sky area fraction of the tracer catalog, used in the volume calculation of `nbar`')
        
        s.add_argument('paintbrush', type=lambda x: x.lower(), choices=['cic', 'tsc'],
            help='the density assignment kernel to use when painting; '
                 'CIC (2nd order) or TSC (3rd order)')
        s.add_argument("dk", type=float,
            help='the spacing of k bins to use; if not provided, '
                 'the fundamental mode of the box is used')
        s.add_argument("kmin", type=float,
            help='the edge of the first `k` bin to use; default is 0')
        s.add_argument('factor_hexadecapole', type=bool, 
            help="use the factored expression for the hexadecapole (ell=4) from "
                 "eq. 27 of Scoccimarro 2015 (1506.02729)")
        s.add_argument('keep_cache', type=bool, 
            help='if `True`, force the data cache to persist while the algorithm instance is valid')
                                
    def run(self):
        """
        Run the algorithm, which computes and returns the power spectrum
        """
        from nbodykit import measurestats
        if self.comm.rank == 0: self.logger.info('importing done')
        
        # explicity store an open stream
        # this prevents the cache from being destroyed while the 
        # algorithm instance is active
        if self.keep_cache:
            self._datacache = self.catalog.data.keep_cache()
            self._rancache = self.catalog.randoms.keep_cache()
        
        # measure
        kws = {'factor_hexadecapole': self.factor_hexadecapole, 'paintbrush':self.paintbrush}
        pm, poles, meta = measurestats.compute_bianchi_poles(self.comm, self.max_ell, self.catalog, self.Nmesh, **kws)
        k3d = pm.k

        # binning in k out to the minimum nyquist frequency
        # (accounting for possibly anisotropic box)
        dk = 2*numpy.pi/pm.BoxSize.min() if self.dk is None else self.dk
        kedges = numpy.arange(self.kmin, numpy.pi*pm.Nmesh/pm.BoxSize.max() + dk/2, dk)

        # project on to 1d k basis
        muedges = numpy.linspace(0, 1, 2, endpoint=True)
        edges = [kedges, muedges]
        poles_final = []
        for p in poles:
            
            # result is (k, mu, power, modes)
            result, _ = measurestats.project_to_basis(pm.comm, k3d, p, edges, symmetric=True)
            poles_final.append(numpy.squeeze(result[2]))
            
        # return (k, poles, modes)
        poles_final = numpy.vstack(poles_final)
        k = numpy.squeeze(result[0])
        modes = numpy.squeeze(result[-1])
        result = k, poles_final, modes

        # compute the metadata to return
        Lx, Ly, Lz = pm.BoxSize
        meta.update({'Lx':Lx, 'Ly':Ly, 'Lz':Lz, 'volume':Lx*Ly*Lz})

        # return all the necessary results
        return kedges, result, meta

    def save(self, output, result):
        """
        Save the power spectrum results to the specified output file

        Parameters
        ----------
        output : str
            the string specifying the file to save
        result : tuple
            the tuple returned by `run()` -- first argument specifies the bin
            edges and the second is a dictionary holding the data results
        """
        from nbodykit.storage import MeasurementStorage
        
        # only the master rank writes
        if self.comm.rank == 0:
            
            kedges, result, meta = result
            k, poles, N = result
            ells = range(0, self.max_ell+1, 2)
            
            # write binned statistic
            args = (",".join(map(str, ells)), output)
            self.logger.info('measurement done; saving ell = %s multipole(s) to %s' %args)
            cols = ['k'] + ['power_%d' %l for l in ells] + ['modes']
            pole_result = [k] + [pole for pole in poles] + [N]
            
            storage = MeasurementStorage.create('1d', output)
            storage.write(kedges, cols, pole_result, **meta)




from .pipeline import RunAlgorithm, add_run_fixture
from . import os, asserts, unittest
from .. import examples_dir

class RunFiberCollisionsAlgorithm(RunAlgorithm):
    run_dir = os.path.join(examples_dir, 'fiber_collisions')
           

@add_run_fixture(__name__, RunFiberCollisionsAlgorithm, 'FiberCollisions')
class TestFiberCollisions(unittest.TestCase):
    param_file  = "test_fc.params"
    output_file = "test_fiber_collisions.hdf5"
    datasources = ['test_bianchi_data.dat']
    
    def test_exit_code(self):
        asserts.test_exit_code(self)
    
    def test_exception(self):
        asserts.test_exception(self) 
    
    def test_result(self):
        asserts.test_pandas_hdf_result(self, "FiberCollisionGroups")




from nbodykit.extensionpoints import Algorithm, DataSource
from nbodykit.distributedarray import GatherArray

import numpy
import logging

def scotts_bin_width(data):
    """
    Return the optimal histogram bin width using Scott's rule
    """
    n = data.size
    sigma = numpy.std(data)
    dx = 3.5 * sigma * 1. / (n ** (1. / 3))
    
    Nbins = numpy.ceil((data.max() - data.min()) * 1. / dx)
    Nbins = max(1, Nbins)
    bins = data.min() + dx * numpy.arange(Nbins + 1)
    return dx, bins

def RedshiftBins(bins):
    """
    The redshift bins to use when computing n(z)
    
    Parameters
    ----------
    bins : int or list of scalars
        If `bins` is a integer, it defines 
        the number of equal-width bins in the given range. If `bins` is a sequence,
        it defines the bin edges, including the rightmost edge, allowing
        for non-uniform bin widths
    """
    if isinstance(bins, list):
        bins = numpy.array(bins)
    elif isinstance(bins, str):
        if 'range' or 'arange':
            try:
                bins = eval(bins)
            except Exception as e:
                raise ValueError("error calling `eval` on `RedshiftBins` value: %s" %str(e))
    elif not isinstance(bins, int):
        raise ValueError("do not understand input ``RedshiftBins`` value of type %s" %type(bins))
    
    return bins
    

class RedshiftHistogramAlgorithm(Algorithm):
    """
    Algorithm to compute the mean number density of as a 
    function of redshift ``n(z)`` for the input ``DataSource``
    """
    plugin_name = "RedshiftHistogram"
    logger = logging.getLogger(plugin_name)

    def __init__(self, datasource, bins=None, fsky=1.0):
        
        # set the cosmology
        self.cosmo = datasource.cosmo
        if self.cosmo is None:
            raise ValueError("`%s` algorithm requires a cosmology" %self.plugin_name)
        
    @classmethod
    def register(cls):
        
        s = cls.schema
        s.description = "compute n(z) from the input DataSource"
        
        s.add_argument("datasource", type=DataSource.from_config,
            help="DataSource with a `Redshift` column to compute n(z) from")
        
        s.add_argument('bins', type=RedshiftBins, 
            help=('the input redshift bins, specified as either as an integer or sequence of floats'))
        s.add_argument('fsky', type=float, 
            help='the sky area fraction, used in the volume calculation for `n(z)`')
                                
    def run(self):
        """
        Run the algorithm, which returns (z, n(z))
        
        Returns
        -------
        zbins : array_like
            the redshift bin edges
        z_cen : array_like
            the center value of each redshift bin
        nz : array_like
            the n(z_cen) value
        """        
        # read the `Redshift` column
        redshift = []
        with self.datasource.open() as stream:
            for [z] in stream.read(['Redshift'], full=False):
                if len(z):
                    if not stream.isdefault('Redshift', z):
                        redshift += list(z)
                    else:
                        raise DataSource.MissingColumn("no ``Redshift`` column in input DataSource")
        
        # gather to root and avoid MPI pickling limits
        redshift = GatherArray(numpy.array(redshift), self.comm, root=0)
        
        # root computes n(z)
        if self.comm.rank == 0:
            
            # use optimal bins via scotts bin width
            if self.bins is None:
                dz, zbins = scotts_bin_width(redshift)
            # use input bins
            elif isinstance(self.bins, int):
                zbins = numpy.linspace(redshift.min(), redshift.max(), self.bins+1, endpoint=True)
            else:
                zbins = self.bins
                
            dig = numpy.searchsorted(zbins, redshift, "right")
            N = numpy.bincount(dig, minlength=len(zbins)+1)[1:-1]
        
            # compute the volume
            R_hi = self.cosmo.comoving_distance(zbins[1:])
            R_lo = self.cosmo.comoving_distance(zbins[:-1])
            volume = (4./3.)*numpy.pi*(R_hi**3 - R_lo**3) * self.fsky
        
            # store the nbar 
            z_cen = 0.5*(zbins[:-1] + zbins[1:])
            nz = 1.*N/volume
        else:
            nz = None; z_cen = None; zbins = None
        
        zbins = self.comm.bcast(zbins)
        z_cen = self.comm.bcast(z_cen)    
        nz = self.comm.bcast(nz)
        
        return zbins, z_cen, nz

    def save(self, output, result):
        """
        Write (z_cen, n(z_cen)) to the specified file

        Parameters
        ----------
        output : str
            the string specifying the file to save
        result : tuple
            the tuple returned by `run()`, which holds (z, nz)
        """
        # only the master rank writes
        if self.comm.rank == 0:
            edges, z_cen, nz = result
            
            with open(output, 'wb') as ff:
                
                ff.write(("# z_min z_max z_cen n(z)\n").encode())
                out = numpy.vstack([edges[:-1], edges[1:], z_cen, nz]).T
                numpy.savetxt(ff, out, fmt='%.6e')


from . import os, cache_dir
from . import results

def assert_numpy_allclose(this, ref, **kws):
    """
    Wrapper around ``numpy.testing.assert_allclose`` that will loop
    over each element of a complex data type, i.e., structured array
    """ 
    from numpy.testing import assert_allclose    
           
    # normal numpy array
    if ref.dtype.names is None:
        assert_allclose(this, ref, **kws)
    # structured array
    else:    
        for name in ref.dtype.names:
            if name not in this.dtype.names:
                raise AssertionError("dtype mismatch in structured data")
            assert_allclose(this[name], ref[name], **kws)

def test_exit_code(self):
    """
    Assert that the exit code is equal to 0
    """
    assert self.run_fixture.exit_code == 0
    
def test_exception(self):
    """
    Assert that no exception occurred, by analyzing the 
    ``stderr`` output
    """
    with open("run.stderr", 'r') as stderr:
        lines = stderr.read()
        assert "Error" not in lines and "Traceback" not in lines
            
def test_result_md5sum(self):
    """
    Assert that the result file from the pipeline run has the same 
    MD5 sum as the reference result file in the cache directory
    """
    this, ref = results.get_result_paths(self.output_file)
    assert results.compute_md5sum(this) == results.compute_md5sum(ref)

def test_pandas_hdf_result(self, dataset, **kws):
    """
    Test that the pandas HDF5 result file is the same as the reference
    result file, by checking that the ``DataFrame`` objects are 
    the same
    
    Parameters
    ----------
    dataset : str
        the name of the HDF5 dataset to load
    **kws : 
        any additional keywords to pass to 
        :func:`~pandas.util.testing.assert_frame_equal`
    """
    from pandas.util.testing import assert_frame_equal
    this, ref = results.get_result_paths(self.output_file)
    
    # the data
    this = results.load_pandas_hdf(this, dataset)
    ref = results.load_pandas_hdf(ref, dataset)
    
    # assert the DataFrames are equal
    assert_frame_equal(this, ref, **kws)
    
def test_hdf_result(self, dataset, rtol=1e-5, atol=1e-8):
    """
    Test that the HDF5 result file is the same as the reference
    result file, using ``h5py``
    
    Parameters
    ----------
    dataset : str
        the name of the HDF5 dataset to load
    rtol : float, optional
        the relative tolerance; default is 1e-5
    atol : float, optional
        the absolute tolerance; default is 1e-8
    """
    this, ref = results.get_result_paths(self.output_file)
    
    # data
    this = results.load_hdf(this, dataset)
    ref = results.load_hdf(ref, dataset)
    
    assert_numpy_allclose(this, ref, rtol=rtol, atol=atol)
        
def test_dataset_result(self, dim, stat, rtol=1e-2, atol=1e-5, skip_imaginary=True):
    """
    Test that the ``DataSet`` result file is the same as the reference
    result file
    
    This loads the `DataSet` and checks that the data
    arrays are ``allclose`` and also checks the meta-data
    
    Parameters
    ----------
    dim : {'1d', '2d'}
        load either a '1d' or '2d' `DataSet`
    stat : {'power', 'corr'}
        load either a power or correlation function measurement
    rtol : float, optional
        the relative tolerance; default is 1e-2
    atol : float, optional
        the absolute tolerance; default is 1e-8
    skip_imaginary : bool, optional
        if `True`, do not check the imaginary components of the data arrays; 
        default is True
    """
    from numpy.testing import assert_array_almost_equal  
    this, ref = results.get_result_paths(self.output_file)
    
    # load datasets
    loader = results.get_dataset_loader(dim, stat)
    this = loader(this); ref = loader(ref)
    
    # check each variable
    for name in ref.variables:
        
        # skip imaginary elements that are very close to zero
        # due to numerical differences
        if skip_imaginary and 'imag' in name:
            continue
        
        if name not in this.variables:
            raise AssertionError("variables name mismatch in ``assert_dataset_result``")
        assert_numpy_allclose(this[name], ref[name], rtol=rtol, atol=atol)
        
    # check the meta-data
    for name in ref.attrs:
        if name not in this.attrs:
            raise AssertionError("meta-data name mismatch in ``assert_dataset_result``")
        assert_array_almost_equal(this.attrs[name], ref.attrs[name])
        
        


from .pipeline import RunAlgorithm, add_run_fixture
from . import os, asserts, unittest
from .. import examples_dir

class RunFOFAlgorithm(RunAlgorithm):
    run_dir = os.path.join(examples_dir, 'fof')
           

@add_run_fixture(__name__, RunFOFAlgorithm, 'FOF')
class TestFastPM(unittest.TestCase):
    param_file  = "test_fastpm.params"
    output_file = "fof_ll0.200_1.0000.hdf5"
    datasources = ['fastpm_1.0000']
    
    def test_exit_code(self):
        asserts.test_exit_code(self)
    
    def test_exception(self):
        asserts.test_exception(self)    
    
    def test_result(self):
        asserts.test_hdf_result(self, "FOFGroups")




from nbodykit.extensionpoints import Algorithm, DataSource
import logging
import numpy

# for output
import h5py
import bigfile

class FOFAlgorithm(Algorithm):
    plugin_name = "FOF"
    
    def __init__(self, datasource, linklength, absolute=False, without_labels=False, nmin=32, calculate_initial_position=False):
        pass
    
    @classmethod
    def register(cls):
        s = cls.schema
        s.description = "Friend of Friend halo finder"
        
        s.add_argument("datasource", type=DataSource.from_config,
            help='`DataSource` objects to run FOF against; '
                 'run `nbkit.py --list-datasources` for all options')
        s.add_argument("linklength", type=float, 
            help='the link length in terms of mean particle sep')
        s.add_argument("absolute", type=bool,
            help='If set, the linking length is in absolute units. '
                 'The default is in relative to mean particle separation.')
        s.add_argument("calculate_initial_position", type=bool,
            help='If set, calculate initial position of halos based on the InitialPosition field of DataSource'
                 )
        s.add_argument("without_labels", type=bool, help='do not store labels')
        s.add_argument("nmin", type=int, help='minimum number of particles in a halo')
        
    def run(self):
        from nbodykit import fof
        # convert to absolute length
        if not self.absolute:
            if not hasattr(self.datasource, 'size'):
                logging.info("Playing DataSource to measure total size. " +
                      "DataSource `%s' shall be fixed to add a size attribute", self.datasource)
                
                with self.datasource.open() as stream:
                    [[junk]] = stream.read(['Position'], full=True)
                    size = self.comm.allreduce(len(junk))
                    del junk
            else:
                size = self.datasource.size
            ll = self.linklength * (self.datasource.BoxSize.prod() / self.datasource.size) ** 0.3333333
        else:
            ll = self.linklength

        labels = fof.fof(self.datasource, ll, self.nmin, self.comm)
        catalog = fof.fof_catalogue(self.datasource, labels, self.comm, self.calculate_initial_position)

        Ntot = self.comm.allreduce(len(labels))
        if self.without_labels:
            return catalog, Ntot
        else:
            return catalog, labels, Ntot

    def save(self, output, data):
        if self.without_labels:
            catalog, Ntot = data
        else:
            catalog, labels, Ntot = data

        if self.comm.rank == 0:
            with h5py.File(output, 'w') as ff:
                # do not create dataset then fill because of
                # https://github.com/h5py/h5py/pull/606

                dataset = ff.create_dataset(
                    name='FOFGroups', data=catalog
                    )
                dataset.attrs['Ntot'] = Ntot
                dataset.attrs['LinkLength'] = self.linklength
                dataset.attrs['BoxSize'] = self.datasource.BoxSize

        if not self.without_labels:
            output = output.replace('.hdf5', '.labels')
            bf = bigfile.BigFileMPI(self.comm, output, create=True)
            with bf.create_from_array("Label", labels, Nfile=(self.comm.size + 7)// 8) as bb:
                bb.attrs['LinkLength'] = self.linklength
                bb.attrs['Ntot'] = Ntot
                bb.attrs['BoxSize'] = self.datasource.BoxSize
        return





from __future__ import print_function

from nbodykit.distributedarray import DistributedArray
from nbodykit.ndarray import equiv_class, replacesorted
from nbodykit import halos

from kdcount import cluster
from pmesh.domain import GridND
import numpy
from mpi4py import MPI

import logging

logger = logging.getLogger('measurestats')

def fof_halo_label(minid, comm, thresh):
    """ 
    Convert minid to sequential labels starting from 0.

    This routine is used to assign halo label to particles with
    the same minid.
    Halos with less than thresh particles are reclassified to 0.

    Parameters
    ----------
    minid : array_like, ('i8')
        The minimum particle id of the halo. All particles of a halo 
        have the same minid
    thresh : int
        halo with less than thresh particles are merged into halo 0
    comm : py:class:`MPI.Comm`
        communicator. since this is a collective operation

    Returns
    -------
    labels : array_like ('i8')
        The new labels of particles. Note that this is ordered
        by the size of halo, with the exception 0 represents all
        particles that are in halos that contain less than thresh particles.
    
    """
    data = numpy.empty(len(minid), dtype=[
            ('origind', 'u8'), 
            ('fofid', 'u8'),
            ])
    # assign origind for recovery of ordering, since
    # we need to work in sorted fofid 
    data['fofid'] = minid
    data['origind'] = numpy.arange(len(data), dtype='u4')
    data['origind'] += sum(comm.allgather(len(data))[:comm.rank]) \
 
    data = DistributedArray(data, comm)

    # first attempt is to assign fofid for each group
    data.sort('fofid')
    label = data['fofid'].unique_labels()
    
    N = label.bincount()
    
    # now eliminate those with less than thresh particles
    small = N.local <= thresh

    Nlocal = label.bincount(local=True)
    # mask == True for particles in small halos
    mask = numpy.repeat(small, Nlocal)
 
    # globally shift halo id by one
    label.local += 1
    label.local[mask] = 0

    data['fofid'].local[:] = label.local[:]
    del label

    data.sort('fofid')

    data['fofid'].local[:] = data['fofid'].unique_labels().local[:]

    data.sort('origind')
    
    label = data['fofid'].local.view('i8').copy()
    
    del data

    Nhalo0 = max(comm.allgather(label.max())) + 1
    Nlocal = numpy.bincount(label, minlength=Nhalo0)
    comm.Allreduce(MPI.IN_PLACE, Nlocal, op=MPI.SUM)

    # sort the labels by halo size
    arg = Nlocal[1:].argsort()[::-1] + 1
    P = numpy.arange(Nhalo0, dtype='i4')
    P[arg] = numpy.arange(len(arg), dtype='i4') + 1
    label = P[label]
        
    return label

def local_fof(layout, pos, boxsize, ll, comm):
    N = len(pos)

    pos = layout.exchange(pos)
    data = cluster.dataset(pos, boxsize=boxsize)
    fof = cluster.fof(data, linking_length=ll, np=0)
    labels = fof.labels
    del fof

    PID = numpy.arange(N, dtype='intp')
    PID += sum(comm.allgather(N)[:comm.rank])

    PID = layout.exchange(PID)
    # initialize global labels
    minid = equiv_class(labels, PID, op=numpy.fmin)[labels]

    return minid

def fof_merge(layout, minid, comm):
    # generate global halo id

    while True:
        # merge, if a particle belongs to several ranks
        # use the global label of the minimal
        minid_new = layout.gather(minid, mode=numpy.fmin)
        minid_new = layout.exchange(minid_new)

        # on my rank, these particles have been merged
        merged = minid_new != minid
        # if no rank has merged any, we are done
        # gl is the global label (albeit with some holes)
        total = comm.allreduce(merged.sum())
            
        if total == 0:
            del minid_new
            break
        old = minid[merged]
        new = minid_new[merged]
        arg = old.argsort()
        new = new[arg]
        old = old[arg]
        replacesorted(minid, old, new, out=minid)

    minid = layout.gather(minid, mode=numpy.fmin)
    return minid

def fof(datasource, linking_length, nmin, comm=MPI.COMM_WORLD, log_level=logging.DEBUG):
    """ Run Friend-of-friend halo finder.

        Friend-of-friend was first used by Davis et al 1985 to define
        halos in hierachical structure formation of cosmological simulations.
        The algorithm is also known as DBSCAN in computer science. 
        The subroutine here implements a parallel version of the FOF. 

        The underlying local FOF algorithm is from `kdcount.cluster`, 
        which is an adaptation of the implementation in Volker Springel's 
        Gadget and Martin White's PM. It could have been done faster.

        Parameters
        ----------
        datasource: DataSource
            datasource; must support Position.
            datasource.BoxSize is used too.
        linking_length: float
            linking length in data units. (Usually Mpc/h).
        nmin: int
            Minimal length (number of particles) of a halo. Features
            with less than nmin particles are considered noise, and
            removed from the catalogue

        comm: MPI.Comm
            The mpi communicator.

        Returns
        -------
        label: array_like
            The halo label of each position. A label of 0 standands for not in any halo.
 
    """
    if log_level is not None: logger.setLevel(log_level)

    np = split_size_3d(comm.size)

    grid = [
        numpy.linspace(0, datasource.BoxSize[0], np[0] + 1, endpoint=True),
        numpy.linspace(0, datasource.BoxSize[1], np[1] + 1, endpoint=True),
        numpy.linspace(0, datasource.BoxSize[2], np[2] + 1, endpoint=True),
    ]
    domain = GridND(grid)

    with datasource.open() as stream:
        [[Position]] = stream.read(['Position'], full=True)

    if comm.rank == 0: logger.info("ll %g. " % linking_length)
    if comm.rank == 0: logger.debug('grid: %s' % str(grid))

    layout = domain.decompose(Position, smoothing=linking_length * 1)

    comm.barrier()
    if comm.rank == 0: logger.info("Starting local fof.")

    minid = local_fof(layout, Position, datasource.BoxSize, linking_length, comm)
    
    comm.barrier()
    if comm.rank == 0: logger.info("Finished local fof.")

    if comm.rank == 0: logger.info("Merged global FOF.")

    minid = fof_merge(layout, minid, comm)
    del layout
    # sort calculate halo catalogue
    label = fof_halo_label(minid, comm, thresh=nmin)

    return label

def fof_catalogue(datasource, label, comm, calculate_initial=False):
    """ Catalogue of FOF groups based on label from a data source

        Friend-of-friend was first used by Davis et al 1985 to define
        halos in hierachical structure formation of cosmological simulations.
        The algorithm is also known as DBSCAN in computer science. 
        The subroutine here implements a parallel version of the FOF. 

        The underlying local FOF algorithm is from `kdcount.cluster`, 
        which is an adaptation of the implementation in Volker Springel's 
        Gadget and Martin White's PM. It could have been done faster.

        Parameters
        ----------
        label : array_like
            halo label of particles from data source. 

        datasource: DataSource
            datasource; must support Position and Velocity.
            datasource.BoxSize is used too.

        comm: MPI.Comm
            The mpi communicator. Must agree with the datasource

        Returns
        -------
        catalogue: array_like
            A 1-d array of type 'Position', 'Velocity', 'Length'. 
            The center mass position and velocity of the FOF halo, and
            Length is the number of particles in a halo. The catalogue is
            sorted such that the most massive halo is first. catalogue[0]
            does not correspond to any halo.
 
    """
    dtype=[
        ('Position', ('f4', 3)),
        ('Velocity', ('f4', 3)),
        ('Length', 'i4')]

    N = halos.count(label, comm=comm)
    
    # explicitly open the DataSource
    with datasource.keep_cache():
    
        with datasource.open() as stream:
            [[Position]] = stream.read(['Position'], full=True)
        Position /= datasource.BoxSize
        hpos = halos.centerofmass(label, Position, boxsize=1.0, comm=comm)
        del Position

        with datasource.open() as stream: 
            [[Velocity]] = stream.read(['Velocity'], full=True)
        Velocity /= datasource.BoxSize

        hvel = halos.centerofmass(label, Velocity, boxsize=None, comm=comm)
        del Velocity

        if calculate_initial:

            dtype.append(('InitialPosition', ('f4', 3)))
        
            with datasource.open() as stream:
                [[Position]] = stream.read(['InitialPosition'], full=True)
            Position /= datasource.BoxSize
            hpos_init = halos.centerofmass(label, Position, boxsize=1.0, comm=comm)
            del Position

    if comm.rank == 0: logger.info("Calculated catalogue %d halos found. " % (len(N) -1 ))
    if comm.rank == 0: logger.info("Length = %s " % N[1:])
    if comm.rank == 0: logger.info("%d particles not in halo" % N[0])

    if comm.rank == 0:
        catalogue = numpy.empty(shape=len(N), dtype=dtype)

        catalogue['Position'] = hpos
        catalogue['Velocity'] = hvel
        catalogue['Length'] = N
        catalogue['Length'][0] = 0
        if calculate_initial:
            catalogue['InitialPosition'] = hpos_init
    else:
        catalogue = numpy.empty(shape=0, dtype=dtype)
        
    return catalogue

def split_size_3d(s):
    """ Split `s` into two integers, 
        a and d, such that a * d == s and a <= d

        returns:  a, d
    """
    a = int(s** 0.33333) + 1
    d = s
    while a > 1:
        if s % a == 0:
            s = s // a
            break
        a = a - 1 
    b = int(s**0.5) + 1
    while b > 1:
        if s % b == 0:
            s = s // b
            break
        b = b - 1
    return a, b, s



import os
import traceback
import logging
import numpy

from mpi4py import MPI
from nbodykit.extensionpoints import set_nbkit_comm

def split_ranks(N_ranks, N, include_all=False):
    """
    Divide the ranks into chunks, attempting to have `N` ranks
    in each chunk. This removes the master (0) rank, such 
    that `N_ranks - 1` ranks are available to be grouped
    
    Parameters
    ----------
    N_ranks : int
        the total number of ranks available
    N : int
        the desired number of ranks per worker
    include_all : bool, optional
        if `True`, then do not force each group to have 
        exactly `N` ranks, instead including the remainder as well;
        default is `False`
    """
    available = list(range(1, N_ranks)) # available ranks to do work    
    total = len(available)
    extra_ranks = total % N
  
    if include_all:
        for i, chunk in enumerate(numpy.array_split(available, total//N)):
            yield i, list(chunk)
    else:
        for i in range(total//N):
            yield i, available[i*N:(i+1)*N]
    
        if extra_ranks and extra_ranks >= N//2:
            remove = extra_ranks % 2 # make it an even number
            ranks = available[-extra_ranks:]
            if remove: ranks = ranks[:-remove]
            if len(ranks):
                yield i+1, ranks
            
def enum(*sequential, **named):
    enums = dict(zip(sequential, range(len(sequential))), **named)
    return type('Enum', (), enums)
    

class TaskManager(object):
    """
    An MPI manager that distributes tasks over a set of MPI processes, 
    using a specified number of independent workers to compute tasks
    
    Given the specified number of independent workers (which compute
    tasks in parallel), the total number of available CPUs will be 
    divided evenly.
    
    The main function is ``compute`` which tasks a list of tasks
    and runs the task function for each item of the list
    """
    logger = logging.getLogger('TaskManager')
    
    def __init__(self, task_function, 
                       cpus_per_worker, 
                       comm=None, 
                       debug=False, 
                       use_all_cpus=False):
        """
        Parameters
        ----------
        task_function : callable
            the task function which takes two arguments: the task iteration
            integer and the `task` value
        cpus_per_worker : int
            the desired number of ranks assigned to each independent
            worker in the pool
        comm : MPI communicator, optional
            the global communicator that will be split so each worker
            has a subset of CPUs available; default is COMM_WORLD
        debug : bool, optional
            if `True`, set the logging level to `DEBUG`, which prints 
            out much more information; default is `False`
        use_all_cpus : bool, optional
            if `True`, use all available CPUs, including the remainder
            if `cpus_per_worker` is not divide the total number of CPUs
            evenly; default is `False`
        """
        if debug:
            self.logger.setLevel(logging.DEBUG)
            
        self.function        = task_function
        self.cpus_per_worker = cpus_per_worker
        self.use_all_cpus    = use_all_cpus
        
        # the main communicator
        self.comm      = MPI.COMM_WORLD if comm is None else comm
        self.rank      = self.comm.rank
        self.size      = self.comm.size
                
        # need at least one
        if self.size == 1:
            raise ValueError("need at least two processes to use a TaskManager")

        # make the sub-communicator
        self._create_subcomm()
        self.status = MPI.Status()
        
        # communication tags
        self.tags = enum('READY', 'DONE', 'EXIT', 'START')

    def _create_subcomm(self):
        """
        Internal function to create the sub-communicator that will be 
        used by each independent worker when running the task function
        
        Notes
        -----
        This function updates the global ``nbodykit`` communicator
        via ``set_nbkit_comm`` so all plugins will be initialized 
        with this sub-communicator
        """
        self.subcomm = None
        chain_ranks = []
        color = 0
        total_ranks = 0
        
        # split the ranks
        for i, ranks in split_ranks(self.size, self.cpus_per_worker, include_all=self.use_all_cpus):
            chain_ranks.append(ranks[0])
            if self.rank in ranks: color = i+1
            total_ranks += len(ranks)
        
        self.workers = i+1 # store the total number of workers
        leftover = (self.size - 1) - total_ranks
        if leftover and self.rank == 0:
            args = (self.cpus_per_worker, self.size-1, leftover)
            self.logger.warning("with `cpus_per_worker` = %d and %d available ranks, %d ranks will do no work" %args)
            
        # crash if we only have one process or one worker
        if self.size <= self.workers:
            args = (self.size, self.workers+1, self.workers)
            raise ValueError("only have %d ranks; need at least %d to use the desired %d workers" %args)
            
        # ranks that will do work have a nonzero color now
        self._valid_worker = color > 0
        
        # split the comm between the workers
        self.subcomm = self.comm.Split(color, 0)
        
        # set the global extension point comm
        set_nbkit_comm(self.subcomm)
        
    def is_master(self):
        """
        Is the current process the master?
        """
        return self.rank == 0

    def is_worker(self):
        """
        Is the current process a valid worker (and thus should wait for
        instructions from the master)
        """
        return self._valid_worker
        
    def wait(self):
        """
        If this isn't the master process, wait for instructions.
        """
        if self.is_master():
            raise RuntimeError("master node told to await jobs")

        # logging info
        if self.subcomm.rank == 0:
            args = (self.rank, MPI.Get_processor_name(), self.subcomm.size)
            self.logger.info("worker master rank is %d on %s with %d processes available" %args)

        # continously loop and wait for instructions
        while True:
            args = None
            tag = -1

            # have the master rank of the subcomm ask for task and then broadcast
            if self.subcomm.rank == 0:
                self.comm.send(None, dest=0, tag=self.tags.READY)
                args = self.comm.recv(source=0, tag=MPI.ANY_TAG, status=self.status)
                tag = self.status.Get_tag()
            
            # bcast to everyone in the worker subcomm
            args  = self.subcomm.bcast(args)
            tag   = self.subcomm.bcast(tag)

            # do the work here
            if tag == self.tags.START:
                result = self.function(*args)
                self.subcomm.Barrier() # wait for everyone in subcomm
                if self.subcomm.rank == 0:
                    self.comm.send([args[0], result], dest=0, tag=self.tags.DONE) # done this task
            elif tag == self.tags.EXIT:
                break

        # wait for everyone in subcomm and exit
        self.subcomm.Barrier()
        if self.subcomm.rank == 0:
            self.comm.send(None, dest=0, tag=self.tags.EXIT) # exiting
            
        # debug logging
        self.logger.debug("rank %d process is done waiting" %self.rank)

    def compute(self, tasks):
        """
        Compute a series of tasks. For each task, the function takes
        the iteration number, followed by the `task` value as the 
        arguments
        
        Parameter
        ---------
        tasks : list
            list of `task` items that will be pickled and set
            to each worker when computing the `ith` task

        Returns
        -------
        results : list
            a list of the return values of the task function for 
            each task
        """
        ntasks = len(tasks)
        results = []

        # catch errors
        try:
    
            # master distributes the tasks and tracks closed workers
            if self.is_master():
                
                # initialize
                task_index = 0
                closed_workers = 0
                
                # logging info
                args = (self.workers, ntasks)
                self.logger.info("master starting with %d worker(s) with %d total tasks" %args)
                
                # loop until all workers have finished with no more tasks
                while closed_workers < self.workers:
                
                    data = self.comm.recv(source=MPI.ANY_SOURCE, tag=MPI.ANY_TAG, status=self.status)
                    source = self.status.Get_source()
                    tag = self.status.Get_tag()
    
                    # worker is ready, so send it a task
                    if tag == self.tags.READY:
                        
                        # still more tasks to compute
                        if task_index < ntasks:
                            this_task = [task_index, tasks[task_index]]
                            self.comm.send(this_task, dest=source, tag=self.tags.START)
                            self.logger.info("sending task `%s` to worker %d" %(str(tasks[task_index]), source))
                            task_index += 1
                        # all tasks sent -- tell worker to exit
                        else:
                            self.comm.send(None, dest=source, tag=self.tags.EXIT)
                    # store the results from finished tasks
                    elif tag == self.tags.DONE:
                        results.append(data)
                        self.logger.debug("received result from worker %d" %source)
                    # track workers that exited
                    elif tag == self.tags.EXIT:
                        closed_workers += 1
                        self.logger.debug("worker %d has exited, closed workers = %d" %(source, closed_workers))
            
            # workers will wait for instructions           
            elif self.is_worker():
                self.wait()
                    
        except Exception as e:
            self.logger.error("an exception has occurred on one of the ranks...all ranks exiting")
            self.logger.error(traceback.format_exc())
            
            # bit of hack that forces mpi4py to exit all ranks
            # see https://groups.google.com/forum/embed/#!topic/mpi4py/RovYzJ8qkbc
            os._exit(1)  
            
        finally:
            # wait and exit
            self.logger.debug("rank %d process finished" %self.rank)
            self.comm.Barrier()
            
            if self.is_master():
                self.logger.info("master is finished; terminating")
                if self.subcomm is not None:
                    self.subcomm.Free()
            
        # return the results in sorted order
        results = self.comm.bcast(results)
        return [r[1] for r in sorted(results, key=lambda r: r[0])]


from nbodykit.extensionpoints import Algorithm, DataSource
from nbodykit import fof
import logging
import numpy
import mpsort
from mpi4py import MPI

logger = logging.getLogger('FiberCollisions')

def RaDecDataSource(d):
    source = DataSource.registry.RaDecRedshift
    d['unit_sphere'] = True
    return source.from_config(d)
            
class FiberCollisionsAlgorithm(Algorithm):
    """
    Run an angular FOF algorithm to determine fiber collision
    groups from an input catalog, and then assign fibers such that
    the maximum amount of object receive a fiber. This amounts
    to determining the following population of objects:
    
        * population 1: 
            the maximal "clean" sample of objects in which each object is not 
            angularly collided with any other object in this subsample
        * population 2:
            the potentially-collided objects; these objects are those
            that are fiber collided + those that have been "resolved"
            due to multiple coverage in tile overlap regions
    
    See Guo et al. 2010 (http://arxiv.org/abs/1111.6598)for further details
    """
    plugin_name = "FiberCollisions"
    
    def __init__(self, datasource, collision_radius=62/60./60., seed=0):
        
        # store collision radius in radians
        self._collision_radius_rad = self.collision_radius * numpy.pi/180.
        if self.comm.rank == 0: 
            logger.info("collision radius in degrees = %.4f" %collision_radius)
            
        # set the seed explicitly for reproducibility
        self.random = numpy.random.RandomState(self.seed)
        
    @classmethod
    def register(cls):

        s = cls.schema
        s.description = "the application of fiber collisions to a galaxy survey"
        
        s.add_argument("datasource", type=RaDecDataSource,
            help='`RaDecRedshift DataSource; run `nbkit.py --list-datasources RaDecRedshift` for details')
        s.add_argument("collision_radius", type=float, 
            help="the size of the angular collision radius (in degrees)")
        s.add_argument("seed", type=int,
            help="seed the random number generator explicitly, for reproducibility")
        
    def run(self):
        """
        Compute the FOF collision groups and assign fibers, such that
        the maximum number of objects receive fibers
        
        Returns
        -------
        result: array_like
            a structured array with 3 fields:
                Label : 
                    the group labels for each object in the input 
                    DataSource; label == 0 objects are not in a group
                Collided : 
                    a flag array specifying which objects are 
                    collided, i.e., do not receive a fiber
                NeighborID : 
                    for those objects that are collided, this 
                    gives the (global) index of the nearest neighbor 
                    on the sky (0-indexed), else it is set to -1
        """
        from nbodykit import fof
        
        # open a persistent cache
        with self.datasource.keep_cache():
            
            # run the angular FoF algorithm to get group labels
            # labels gives the global group ID corresponding to each object in Position 
            # on this rank
            labels = fof.fof(self.datasource, self._collision_radius_rad, 1, comm=self.comm)

            # assign the fibers (in parallel)
            collided, neighbors = self._assign_fibers(labels)
    
        # all reduce to get summary statistics
        N_pop1 = self.comm.allreduce((collided^1).sum())
        N_pop2 = self.comm.allreduce((collided).sum())
        f = N_pop2 * 1. / (N_pop1 + N_pop2)

        # print out some info
        if self.comm.rank == 0:

            logger.info("population 1 (clean) size = %d" %N_pop1)
            logger.info("population 2 (collided) size = %d" %N_pop2)
            logger.info("collision fraction = %.4f" %f)

        # return a structured array
        d = list(zip(['Label', 'Collided', 'NeighborID'], [labels, collided, neighbors]))
        dtype = [(col, x.dtype) for col, x in d]
        result = numpy.empty(len(labels), dtype=dtype)
        for col, x in d: result[col] = x
        return result

    def _assign_fibers(self, Label):
        """
        Initernal function to divide the data by collision group 
        across ranks and assign fibers, such that the minimum
        number of objects are collided out of the survey
        """
        comm = self.comm
        mask = Label != 0
        PIG = numpy.empty(mask.sum(), dtype=[
                ('Position', ('f4', 3)),  
                ('Label', ('i4')), 
                ('Rank', ('i4')), 
                ('Index', ('i4')),
                ('Collided', ('i4')),
                ('NeighborID', ('i4'))
                ])
        PIG['Label'] = Label[mask]
        size = len(Label)
        size = comm.allgather(size)
        Ntot = sum(size)
        offset = sum(size[:comm.rank])
        PIG['Index'] = offset + numpy.where(mask == True)[0]
        del Label
        
        with self.datasource.open() as stream:
            [[Position]] = stream.read(['Position'], full=True)
        PIG['Position'] = Position[mask]
        del Position
        Ntot = comm.allreduce(len(mask))
        Nhalo = comm.allreduce(
            PIG['Label'].max() if len(PIG['Label']) > 0 else 0, op=MPI.MAX) + 1

        # now count number of particles per halo
        PIG['Rank'] = PIG['Label'] % comm.size
        cnt = numpy.bincount(PIG['Rank'], minlength=comm.size)
        Nlocal = comm.allreduce(cnt)[comm.rank]

        # sort by rank and then label
        PIG2 = numpy.empty(Nlocal, PIG.dtype)
        mpsort.sort(PIG, orderby='Rank', out=PIG2)
        assert (PIG2['Rank'] == comm.rank).all()
        PIG2.sort(order=['Label'])
        
        if self.comm.rank == 0:
            logger.info('total number of collision groups = %d', Nhalo-1)
            logger.info("Started fiber assignment")

        # loop over unique group ids
        for group_id in numpy.unique(PIG2['Label']):
            start = PIG2['Label'].searchsorted(group_id, side='left')
            end = PIG2['Label'].searchsorted(group_id, side='right')
            N = end-start
            assert(PIG2['Label'][start:end] == group_id).all()
            
            # pairs (random selection)
            if N == 2:
                which = self.random.choice([0,1])
                indices = [start+which, start+(which^1)]
                PIG2['Collided'][indices] = [1, 0]
                PIG2['NeighborID'][indices] = [PIG2['Index'][start+(which^1)], -1]
            # multiplets (minimize collidedness)
            elif N > 2:
                collided, nearest = self._assign_multiplets(PIG2['Position'][start:end])
                PIG2['Collided'][start:end] = collided[:]
                PIG2['NeighborID'][start:end] = -1
                PIG2['NeighborID'][start:end][collided==1] = PIG2['Index'][start+nearest][:]

        if self.comm.rank == 0: logger.info("Finished fiber assignment")
    
        # return to the order specified by the global unique index
        mpsort.sort(PIG2, orderby='Index', out=PIG)
        
        # return arrays including the objects not in any groups
        collided = numpy.zeros(size[comm.rank], dtype='i4')
        collided[mask] = PIG['Collided'][:]
        neighbors = numpy.zeros(size[comm.rank], dtype='i4') - 1
        neighbors[mask] = PIG['NeighborID'][:]

        del PIG
        return collided, neighbors
    
    def _assign_multiplets(self, Position):
        """
        Internal function to assign the maximal amount of fibers 
        in collision groups of size N > 2
        """
        from scipy.spatial.distance import pdist, squareform
        
        def count(slice, n):
            return n[numpy.nonzero(slice)[0]].sum()
        
        # first shuffle the member ids, so we select random element when tied
        N = len(Position)
        group_ids = list(range(N))
    
        collided_ids = []
        while len(group_ids) > 1:
       
            # compute dists and find where dists < collision radius
            dists = squareform(pdist(Position[group_ids], metric='euclidean'))
            numpy.fill_diagonal(dists, numpy.inf) # ignore self-pairs
            collisions = dists <= self._collision_radius_rad
            
            # total # of collisions for each group member
            n_collisions = numpy.sum(collisions, axis=0)
            
            # total # of collisions for those objects that collide with each group member
            n_other = numpy.apply_along_axis(count, 0, collisions, n_collisions)
            
            # remove object that has most # of collisions 
            # and those colliding objects have least # of collisions
            idx = numpy.where(n_collisions == n_collisions.max())[0]
            ii = self.random.choice(numpy.where(n_other[idx] == n_other[idx].min())[0])
            collided_index = idx[ii]  
    
            # make the collided galaxy and remove from group
            collided_id = group_ids.pop(collided_index)
        
            # only make this a collided object if its n_collisions > 0
            # if n_collisions = 0, then the object can get a fiber for free
            if n_collisions[collided_index] > 0:
                collided_ids.append(collided_id)

        # compute the nearest neighbors
        neighbor_ids = []
        group_indices = list(range(N))
        dists = squareform(pdist(Position, metric='euclidean'))
        uncollided = [i for i in group_indices if i not in collided_ids]
        
        for i in sorted(collided_ids):
            neighbor = uncollided[dists[i][uncollided].argmin()]
            neighbor_ids.append(neighbor)
            
        collided = numpy.zeros(N)
        collided[collided_ids] = 1
        return collided, neighbor_ids
            
    def save(self, output, result):
        """
        Write the `Label`, `Collided`, and `NeighborID` arrays
        as a Pandas DataFrame to an HDF file, with key `FiberCollisonGroups`
        """
        import pandas as pd
        import os

        # gather the result to root and output
        result = self.comm.gather(result, root=0)
        
        if self.comm.rank == 0:
            
            # enforce a default extension
            _, ext = os.path.splitext(output)
            if 'hdf' not in ext: output += '.hdf5'
            
            logger.info("saving (Label, Collided, NeighborID) as Pandas HDF with name %s" %output)
        
            result = numpy.concatenate(result, axis=0)
            df = pd.DataFrame.from_records(result)
            df.to_hdf(output, 'FiberCollisionGroups')




from nbodykit.extensionpoints import DataSource
from nbodykit import files 
import numpy

class GadgetDataSource(DataSource):
    """
    A DataSource to read a flavor of Gadget 2 files (experimental)
    """
    plugin_name = "Gadget"
    
    def __init__(self, path, BoxSize, ptype=[], posdtype='f4', veldtype='f4', 
                    iddtype='u8', massdtype='f4', rsd=None, bunchsize=4*1024*1024):
        pass
    
    @classmethod
    def register(cls):
        
        s = cls.schema
        s.description = "read a flavor of Gadget 2 files (experimental)"
        
        s.add_argument("path", type=str, help="the file path to load the data from")
        s.add_argument("BoxSize", type=cls.BoxSizeParser,
            help="the size of the isotropic box, or the sizes of the 3 box dimensions")
        s.add_argument("ptype", nargs='*', type=int, help="the particle types to use")
        s.add_argument("posdtype", type=str, help="dtype of position")
        s.add_argument("veldtype", type=str, help="dtype of velocity")
        s.add_argument("iddtype", type=str, help="dtype of id")
        s.add_argument("massdtype", type=str, help="dtype of mass")
        s.add_argument("rsd", choices="xyz", type=str, help="direction to do redshift distortion")
        s.add_argument("bunchsize", type=int, help="number of particles to read per rank in a bunch")

    def parallel_read(self, columns, full=False):
        """ read data in parallel. if Full is True, neglect bunchsize. """
        Ntot = 0
        # avoid reading Velocity if RSD is not requested.
        # this is only needed for large data like a TPMSnapshot
        # for small Pandas reader etc it doesn't take time to
        # read velocity

        if self.rsd is not None:
            newcolumns = set(columns + ['Velocity'])
        else:
            newcolumns = set(columns)

        bunchsize = self.bunchsize
        if full: bunchsize = -1
        if full and len(self.ptype) > 1:
            raise ValueError("cannot read multple ptype in a full load")
        for ptype in self.ptype:
            args = dict(ptype=ptype,
                 posdtype=self.posdtype,
                 veldtype=self.veldtype,
                 massdtype=self.massdtype,
                 iddtype=self.iddtype)

            if self.comm.rank == 0:
                datastorage = files.DataStorage(self.path,
                        files.GadgetSnapshotFile, args)
                f0 = files.GadgetSnapshotFile(self.path, 0, args)
                boxsize = f0.header['boxsize']
            else:
                datastorage = None
                boxsize = None
            boxsize = self.comm.bcast(boxsize)
            datastorage = self.comm.bcast(datastorage)

            for round, P in enumerate(
                    datastorage.iter(comm=self.comm, 
                        columns=newcolumns, bunchsize=bunchsize)):
                P = dict(zip(newcolumns, P))
                if 'Position' in P:
                    P['Position'] /= boxsize
                    P['Position'] *= self.BoxSize
                if 'Velocity' in P:
                    raise KeyError('Velocity is not yet supported')

                if self.rsd is not None:
                    dir = "xyz".index(self.rsd)
                    P['Position'][:, dir] += P['Velocity'][:, dir]
                    P['Position'][:, dir] %= self.BoxSize[dir]

                yield [P.get(key, None) for key in columns]


class GadgetGroupTabDataSource(DataSource):
    """
    A DataSource to read a flavor of Gadget 2 FOF catalogs (experimental)
    """
    plugin_name = "GadgetGroupTab"
    
    def __init__(self, path, BoxSize, mpch=1000., posdtype='f4', veldtype='f4', 
                    massdtype='f4', rsd=None, bunchsize=4*1024*1024):
        pass
    
    @classmethod
    def register(cls):
        
        s = cls.schema
        s.description = "read a flavor of Gadget 2 FOF catalogs (experimental)"
        
        s.add_argument("path", type=str, help="the file path to load the data from")
        s.add_argument("BoxSize", type=cls.BoxSizeParser,
            help="the size of the isotropic box, or the sizes of the 3 box dimensions")
        s.add_argument("mpch", type=float, help="Mpc/h in code unit")
        s.add_argument("posdtype", help="dtype of position")
        s.add_argument("veldtype", help="dtype of velocity")
        s.add_argument("massdtype", help="dtype of mass")
        s.add_argument("rsd", choices="xyz", help="direction to do redshift distortion")
        s.add_argument("bunchsize", type=int, 
            help="number of particles to read per rank in a bunch")
    
    def read(self, columns, full=False):
        """ read data in parallel. if Full is True, neglect bunchsize. """
        Ntot = 0
        # avoid reading Velocity if RSD is not requested.
        # this is only needed for large data like a TPMSnapshot
        # for small Pandas reader etc it doesn't take time to
        # read velocity

        if self.rsd is not None:
            newcolumns = set(columns + ['Velocity'])
        else:
            newcolumns = set(columns)

        bunchsize = self.bunchsize
        if full: bunchsize = -1

        args = dict(posdtype=self.posdtype,
             veldtype=self.veldtype,
             massdtype=self.massdtype,
             iddtype=self.iddtype)

        if self.comm.rank == 0:
            datastorage = files.DataStorage(self.path,
                    files.GadgetGroupTabFile, args)
        else:
            datastorage = None
        datastorage = self.comm.bcast(datastorage)

        for round, P in enumerate(
                datastorage.iter(comm=self.comm, 
                    columns=newcolumns, bunchsize=bunchsize)):
            P = dict(zip(newcolumns, P))
            if 'Position' in P:
                P['Position'] /= self.mpch
            if 'Velocity' in P:
                raise KeyError('Velocity is not yet supported')

            if self.rsd is not None:
                dir = "xyz".index(self.rsd)
                P['Position'][:, dir] += P['Velocity'][:, dir]
                P['Position'][:, dir] %= self.BoxSize[dir]

            yield [P[key] for key in columns]



from . import os, cache_dir
from .. import pkg_dir

def get_result_paths(output_file):
    """
    Return the full reference paths for a given output file name.
    
    Parameters
    ----------
    output_file : str
        the output file name
    
    Returns
    -------
    this : str
        the full output file path in the `nbodykit/examples/output` directory
    ref : str
        the output file path in the '~/.nbodykit/results' directory
    """
    this = os.path.join(pkg_dir, 'examples', 'output', output_file)
    ref = os.path.join(os.path.expanduser(cache_dir), 'results', output_file)
    return this, ref
        
def compute_md5sum(path):
    """
    Return the md5 hash for the input path. If the input path is
    a directory, return the concatenatation of all files in the directory
    
    Parameters
    ----------
    path : str
        the name of a file or directory path
    
    Returns
    -------
    str : 
        the md5 sum hash string
    """    
    from pytest_pipeline import utils
    
    if not os.path.exists(path):
        raise ValueError("no result file located at '%s'" %path)
        
    if os.path.isdir(path):
        paths_to_hash = []
        for root, directories, filenames in os.walk(path):
            for filename in filenames:
                paths_to_hash.append(os.path.join(root,filename))
        
        return "".join(utils.file_md5sum(f) for f in sorted(paths_to_hash))
        
    else:
        return utils.file_md5sum(path) 
        
def load_pandas_hdf(filename, dataset):
    """
    Load a pandas HDF file, returning the ``DataFrame`` object
    
    Parameters
    ----------
    dataset : str
        the name of the HDF5 dataset to load
    
    Returns
    -------
    pandas.DataFrame : 
        the loaded data frame
    """
    from pandas import read_hdf
    return read_hdf(filename, dataset)
    
def load_hdf(filename, dataset):
    """
    Load a HDF5 file using ``h5py``, returning a ``numpy`` array
    
    Parameters
    ----------
    dataset : str
        the name of the HDF5 dataset to load
    
    Returns
    -------
    array_like : 
        the loaded data
    """
    import h5py
    return h5py.File(filename, mode='r')[dataset][...]
    
def get_dataset_loader(dim, stat):
    """
    Load a ``Dataset`` object from file
    
    Parameters
    ----------
    dim : {'1d', '2d'}
        load either a '1d' or '2d' `DataSet`
    stat : {'power', 'corr'}
        load either a power or correlation function measurement
    
    Returns
    -------
    callable : 
        a function that takes the file name as its only argument
        and returns a `DataSet`
    """
    from nbodykit import files, dataset
    
    if stat not in ['power', 'corr']:
        raise ValueError("``stat`` in `load_dataset` should be one of ['power', 'corr']")
    if dim not in ['1d', '2d']:
        raise ValueError("``dim`` in `load_dataset` should be one of ['1d', '2d']")
       
    if dim == '2d':
        reader = files.Read2DPlainText
        cls = dataset.Power2dDataSet if stat == 'power' else dataset.Corr2dDataSet
    else:
        reader = files.Read1DPlainText
        cls = dataset.Power1dDataSet if stat == 'power' else dataset.Corr1dDataSet
    
    return lambda f: cls.from_nbkit(*reader(f))

    


import numpy
import logging

__all__ = ['read', 'TPMSnapshotFile', 'SnapshotFile']

class StripeFile:
    @classmethod
    def enum(filetype, basename, args):
        """
        Iterate over all files of the type
        """
        i = 0
        while True:
            try:
                yield filetype(basename, i, args)
            except IOError as e:
                # if file does not open properly, we are done
                break
            i = i + 1
    
    column_names = set([
       'Position', 
       'Mass', 
       'ID',
       'Velocity',
       'Label',
    ])
    def read(self, column, mystart, myend):
        """
        Read a property column of particles

        Parameters
        ----------
        mystart   : int
            offset to start reading within this file. (inclusive)
        myend     : int
            offset to end reading within this file. (exclusive)

        Returns
        -------
        data : array_like (myend - mystart)
            data in unspecified units.

        """
        return NotImplementedError

    def write(self, column, mystart, data):
        return NotImplementedError

    def readat(self, offset, nitem, dtype):
        with open(self.filename, 'rb') as ff:
            ff.seek(offset, 0)
            return numpy.fromfile(ff, count=nitem, dtype=dtype)

    def writeat(self, offset, data):
        with open(self.filename, 'rb+') as ff:
            ff.seek(offset, 0)
            return ff.tofile(ff)

class DataStorage(object):
    """ DataStorage provides a continuous view of 
        across several files.
    """
    def __init__(self, path, filetype, filetype_args={}):
        """ filetype must be of a subclass of StripeFile """
        self.npart = numpy.array(
            [ff.npart for ff in filetype.enum(path, filetype_args)],
            dtype='i8')

        self.path = path
        self.filetype = filetype
        self.filetype_args = filetype_args
        if len(self.npart) == 0:
            raise IOError("No files were found under `%s`" % path)

    @classmethod
    def create(kls, path, filetype, npart, filetype_args={}):
        """ create a striped file. 
            npart is a list of npart for each file 
        """
        for fid, npart1 in enumerate(npart):
            filetype.create(path, fid, npart1, filetype_args)

        self = kls(path, filetype, filetype_args)
        return self

    def get_file(self, i):
        return self.filetype(self.path, i, self.filetype_args)

    def read(self, column, start, end):

        NpartPerFile = self.npart
        NpartCumFile = numpy.concatenate([[0], numpy.cumsum(self.npart)])
        data = []
        # add an empty item
        ff = self.filetype(self.path, 0, self.filetype_args)
        data.append(ff.read(column, 0, 0))

        for i in range(len(NpartPerFile)):
            if end <= NpartCumFile[i]: continue
            if start >= NpartCumFile[i+1]: continue
            # find the intersection in this file
            mystart = max(start - NpartCumFile[i], 0)
            myend = min(end - NpartCumFile[i], NpartPerFile[i])

            ff = self.filetype(self.path, i, self.filetype_args)
            data.append(ff.read(column, mystart, myend))
        
        data = numpy.concatenate(data, axis=0)
        if not len(data): data = None
        return data

    def write(self, column, start, data):
        """this function provides a continuous view of multiple files"""

        NpartPerFile = self.npart
        NpartCumFile = numpy.concatenate([[0], numpy.cumsum(self.npart)])
        ff = self.filetype(self.path, 0, self.filetype_args)
        end = start + len(data)
        offset = 0
        for i in range(len(NpartPerFile)):
            if end <= NpartCumFile[i]: continue
            if start >= NpartCumFile[i+1]: continue
            # find the intersection in this file
            mystart = max(start - NpartCumFile[i], 0)
            myend = min(end - NpartCumFile[i], NpartPerFile[i])

            ff = self.filetype(self.path, i, self.filetype_args)
            ff.write(column, mystart, data[offset:offset + myend - mystart])
            offset += myend - mystart
    
        return 

    def iter(self, columns, bunchsize, comm):
        """
        Parallel reading. This is a generator function and a collective 
        operation.

        Use a for loop. For example
        .. code:: python
            
            for i, P in enumerate(read(comm, 'snapshot', TPMSnapshotFile)):
                ....
                # process P

        Parameters
        ----------
        comm  : :py:class:`MPI.Comm`
            Communicator

        bunchsize : int
            Number of particles to read per rank in each iteration.
            if < 0, all particles are read in one iteration

        Yields
        ------
        data in columns, 

        """
        
        Ntot = self.npart.sum()

        mystart = comm.rank * Ntot // comm.size
        myend = (comm.rank + 1) * Ntot // comm.size

        if bunchsize < 0:
            # set to a sufficiently large number.
            bunchsize = int(Ntot)

        Nchunk = 0
        for i in range(mystart, myend, bunchsize):
            Nchunk += 1
        
        # ensure every rank yields the same number of times
        # for decompose is a collective operation.

        Nchunk = max(comm.allgather(Nchunk))
        for i in range(Nchunk):
            a, b, c = slice(mystart + i * bunchsize, 
                            mystart + (i +1)* bunchsize)\
                        .indices(myend) 
            data = []
            for column in columns:
                data.append(self.read(column, a, b))

            # FIXME: we need to fix this ugly thing
            #print comm.allreduce(P['Position'].max(), op=MPI.MAX)
            #print comm.allreduce(P['Position'].min(), op=MPI.MIN)
            yield data
            i = i + bunchsize
            



from nbodykit.extensionpoints import Algorithm, DataSource
import logging
import numpy

# for output
import h5py
import mpsort
from mpi4py import MPI
from kdcount import cluster

class FOF6DAlgorithm(Algorithm):
    plugin_name = "FOF6D"

    def __init__(self, datasource, halolabel, linklength=0.078, vfactor=0.368, nmin=32):
        pass

    @classmethod
    def register(cls):

        s = cls.schema
        s.description = "finding subhalos from FOF groups; a variant of FOF6D"
        
        s.add_argument("datasource", type=DataSource.from_config,
            help='`DataSource` objects to run FOF against; ' 
                 'run `nbkit.py --list-datasources` for all options')
        s.add_argument("halolabel", type=DataSource.from_config,
            help='data source for the halo label files; column name is Label')
        s.add_argument("linklength", type=float, help='the linking length')
        s.add_argument("vfactor", type=float,
               help='velocity linking length in units of 1d velocity dispersion.')
        s.add_argument("nmin", type=int, help='minimum number of particles in a halo')

    def run(self):
        comm = self.comm
        offset = 0
        
        with self.halolabel.open() as stream:
            [[Label]] = stream.read(['Label'], full=True)
        mask = Label != 0
        PIG = numpy.empty(mask.sum(), dtype=[
                ('Position', ('f4', 3)), 
                ('Velocity', ('f4', 3)), 
                ('Label', ('i4')), 
                ('Rank', ('i4')), 
                ])
        PIG['Label'] = Label[mask]
        del Label
        with self.datasource.open() as stream:
            [[Position]] = stream.read(['Position'], full=True)
            PIG['Position'] = Position[mask]
            del Position
            [[Velocity]] = stream.read(['Velocity'], full=True)
            PIG['Velocity'] = Velocity[mask]
            del Velocity
     
        Ntot = comm.allreduce(len(mask))
        del mask

        Nhalo = comm.allreduce(
            PIG['Label'].max() if len(PIG['Label']) > 0 else 0, op=MPI.MAX) + 1

        # now count number of particles per halo
        PIG['Rank'] = PIG['Label'] % comm.size

        Nlocal = comm.allreduce(
                    numpy.bincount(PIG['Rank'], minlength=comm.size)
                 )[comm.rank]

        PIG2 = numpy.empty(Nlocal, PIG.dtype)

        mpsort.sort(PIG, orderby='Rank', out=PIG2)
        del PIG

        assert (PIG2['Rank'] == comm.rank).all()

        PIG2.sort(order=['Label'])

        logging.info('halos = %d', Nhalo)
        cat = []
        for haloid in numpy.unique(PIG2['Label']):
            hstart = PIG2['Label'].searchsorted(haloid, side='left')
            hend = PIG2['Label'].searchsorted(haloid, side='right')
            if hend - hstart < self.nmin: continue
            assert(PIG2['Label'][hstart:hend] == haloid).all()
            cat.append(
                subfof(
                    PIG2['Position'][hstart:hend], 
                    PIG2['Velocity'][hstart:hend], 
                    self.linklength * (self.datasource.BoxSize.prod() / Ntot) ** 0.3333, 
                    self.vfactor, haloid, Ntot, self.datasource.BoxSize))
        cat = numpy.concatenate(cat, axis=0)
        return cat, Ntot

    def save(self, output, data):
        comm = self.comm
        cat, Ntot = data
        cat = comm.gather(cat)

        if comm.rank == 0:
            cat = numpy.concatenate(cat, axis=0)
            with h5py.File(output, mode='w') as f:
                dataset = f.create_dataset('Subhalos', data=cat)
                dataset.attrs['LinkingLength'] = self.linklength
                dataset.attrs['VFactor'] = self.vfactor
                dataset.attrs['Ntot'] = Ntot
                dataset.attrs['BoxSize'] = self.datasource.BoxSize

def subfof(pos, vel, ll, vfactor, haloid, Ntot, boxsize):
    nbar = Ntot / boxsize.prod()
    first = pos[0].copy()
    pos -= first
    pos /= boxsize
    pos[pos > 0.5]  -= 1.0 
    pos[pos < -0.5] += 1.0 
    pos *= boxsize
    pos += first

    oldvel = vel.copy()
    vmean = vel.mean(axis=0, dtype='f8')
    vel -= vmean
    sigma_1d = (vel** 2).mean(dtype='f8') ** 0.5
    vel /= (vfactor * sigma_1d)
    vel *= ll
    data = numpy.concatenate(( pos, vel), axis=1)
    #data = pos

    data = cluster.dataset(data)
    Nsub = 0
    thresh = 80
    fof = cluster.fof(data, linking_length=ll, np=0)

    while Nsub == 0 and thresh > 1:
        # reducing the threshold till we find something..
        Nsub = (fof.length > thresh).sum()
        thresh *= 0.9
    # if nothing is found then assume this FOF group is a fluke.
 
    output = numpy.empty(Nsub, dtype=[
        ('Position', ('f4', 3)),
        ('Velocity', ('f4', 3)),
        ('LinkingLength', 'f4'),
        ('R200', 'f4'),
        ('R500', 'f4'),
        ('R1200', 'f4'),
        ('R2400', 'f4'),
        ('R6000', 'f4'),
        ('Length', 'i4'),
        ('HaloID', 'i4'),
        ])

    output['Position'][...] = fof.center()[:Nsub, :3]
    output['Length'][...] = fof.length[:Nsub]
    output['HaloID'][...] = haloid
    output['LinkingLength'][...] = ll

    for i in range(3):
        output['Velocity'][..., i] = fof.sum(oldvel[:, i])[:Nsub] / output['Length']

    del fof
    del data
    data = cluster.dataset(pos)

    for i in range(Nsub):
        center = output['Position'][i] 
        rmax = (((pos - center) ** 2).sum(axis=-1) ** 0.5).max()
        r1 = rmax
        output['R200'][i] = so(center, data, r1, nbar, 200.)
        output['R500'][i] = so(center, data, r1, nbar, 500.)
        output['R1200'][i] = so(center, data, output['R200'][i] * 0.5, nbar, 1200.)
        output['R2400'][i] = so(center, data, output['R1200'][i] * 0.5, nbar, 2400.)
        output['R6000'][i] = so(center, data, output['R2400'][i] * 0.5, nbar, 6000.)
    return output

def so(center, data, r1, nbar, thresh=200):
    center = numpy.array([center])
    dcenter = cluster.dataset(center)

    def delta(r):
        if r < 1e-7:
            raise StopIteration
        N = data.tree.root.count(dcenter.tree.root, r)
        n = N / (4. / 3. * numpy.pi * r ** 3)
        return 1.0 * n / nbar - 1
     
    try:
        d1 = delta(r1)
        while d1 > thresh:
            r1 *= 1.4
            d1 = delta(r1)
        # d1 < 200
        r2 = r1
        d2 = d1
        while d2 < thresh:
            r2 *= 0.7
            d2 = delta(r2)
        # d2 > 200

        while True:
            r = (r1 * r2) ** 0.5
            d = delta(r)
            x = (d - thresh)
            if x > 0.1:
                r2 = r
            elif x < -0.1:
                r1 = r
            else:
                return r

    except StopIteration:
        return numpy.nan



from nbodykit.extensionpoints import Algorithm
import logging
import numpy

# for output
import h5py
import bigfile
import mpsort

from mpi4py import MPI

import nbodykit
from pmesh.particlemesh import ParticleMesh
from nbodykit.extensionpoints import DataSource, Painter


class Subsample(Algorithm):
    plugin_name = "Subsample"
    
    def __init__(self, datasource, Nmesh, seed=12345, ratio=0.01, smoothing=None, format='hdf5'):
        pass

    @classmethod
    def register(cls):
        
        s = cls.schema
        s.description = "create a subsample from a DataSource, and evaluate density \n"
        s.description += "(1 + delta) smoothed at the given scale"
        
        s.add_argument("datasource", type=DataSource.from_config,
            help="the DataSource to read; run `nbkit.py --list-datasources` for all options")
        s.add_argument("Nmesh", type=int, help='the size of FFT mesh for painting')
        s.add_argument("seed", type=int, help='the random seed')
        s.add_argument("ratio", type=float, help='the fraction of particles to keep')
        s.add_argument("smoothing", type=float,
                help='the smoothing length in distance units. '
                      'It has to be greater than the mesh resolution. '
                      'Otherwise the code will die. Default is the mesh resolution.')
        # this is for output..
        s.add_argument("format", choices=['hdf5', 'mwhite'], help='the format of the output')

    def run(self):
        
        pm = ParticleMesh(self.datasource.BoxSize, self.Nmesh, dtype='f4', comm=self.comm)
        if self.smoothing is None:
            self.smoothing = self.datasource.BoxSize[0] / self.Nmesh
        elif (self.datasource.BoxSize / self.Nmesh > self.smoothing).any():
            raise ValueError("smoothing is too small")
     
        def Smoothing(pm, complex):
            k = pm.k
            k2 = 0
            for ki in k:
                ki2 = ki ** 2
                complex *= numpy.exp(-0.5 * ki2 * self.smoothing ** 2)

        def NormalizeDC(pm, complex):
            """ removes the DC amplitude. This effectively
                divides by the mean
            """
            w = pm.w
            comm = pm.comm
            ind = []
            value = 0.0
            found = True
            for wi in w:
                if (wi != 0).all():
                    found = False
                    break
                ind.append((wi == 0).nonzero()[0][0])
            if found:
                ind = tuple(ind)
                value = numpy.abs(complex[ind])
            value = comm.allreduce(value, MPI.SUM)
            complex[:] /= value
            
        # open the datasource and keep the cache
        with self.datasource.keep_cache():
     
            painter = Painter.create("DefaultPainter")
            painter.paint(pm, self.datasource)
            pm.r2c()
        
            pm.transfer([Smoothing, NormalizeDC])
            pm.c2r()
            columns = ['Position', 'ID', 'Velocity']
            rng = numpy.random.RandomState(self.Nmesh)
            seedtable = rng.randint(1024*1024*1024, size=self.comm.size)
            rngtable = [numpy.random.RandomState(seed) for seed in seedtable]

            dtype = numpy.dtype([
                    ('Position', ('f4', 3)),
                    ('Velocity', ('f4', 3)),
                    ('ID', 'u8'),
                    ('Density', 'f4'),
                    ]) 

            subsample = [numpy.empty(0, dtype=dtype)]

            with self.datasource.open() as stream:
                for Position, ID, Velocity in stream.read(columns):
                    u = rngtable[self.comm.rank].uniform(size=len(ID))
                    keep = u < self.ratio
                    Nkeep = keep.sum()
                    if Nkeep == 0: continue 
                    data = numpy.empty(Nkeep, dtype=dtype)
                    data['Position'][:] = Position[keep]
                    data['Velocity'][:] = Velocity[keep]       
                    data['ID'][:] = ID[keep] 

                    layout = pm.decompose(data['Position'])
                    pos1 = layout.exchange(data['Position'])
                    density1 = pm.readout(pos1)
                    density = layout.gather(density1)

                    # normalize the position after reading out density!
                    data['Position'][:] /= self.datasource.BoxSize
                    data['Velocity'][:] /= self.datasource.BoxSize
                    data['Density'][:] = density
                    subsample.append(data)
        
        subsample = numpy.concatenate(subsample)
        mpsort.sort(subsample, orderby='ID')
        
        return subsample

    def save(self, output, data):
        if self.format == 'mwhite':
            self.write_mwhite_subsample(data, output)
        else:
            self.write_hdf5(data, output)

    def write_hdf5(self, subsample, output):

        size = self.comm.allreduce(len(subsample))
        offset = sum(self.comm.allgather(len(subsample))[:self.comm.rank])

        if self.comm.rank == 0:
            with h5py.File(output, 'w') as ff:
                dataset = ff.create_dataset(name='Subsample',
                        dtype=subsample.dtype, shape=(size,))
                dataset.attrs['Ratio'] = self.ratio
                dataset.attrs['CommSize'] = self.comm.size 
                dataset.attrs['Seed'] = self.seed
                dataset.attrs['Smoothing'] = self.smoothing
                dataset.attrs['Nmesh'] = self.Nmesh
                dataset.attrs['Original'] = self.datasource.string
                dataset.attrs['BoxSize'] = self.datasource.BoxSize

        for i in range(self.comm.size):
            self.comm.barrier()
            if i != self.comm.rank: continue
                 
            with h5py.File(output, 'r+') as ff:
                dataset = ff['Subsample']
                dataset[offset:len(subsample) + offset] = subsample

    def write_mwhite_subsample(self, subsample, output):
        size = self.comm.allreduce(len(subsample))
        offset = sum(self.comm.allgather(len(subsample))[:self.comm.rank])

        if self.comm.rank == 0:
            with open(output, 'wb') as ff:
                dtype = numpy.dtype([
                        ('eflag', 'int32'),
                        ('hsize', 'int32'),
                        ('npart', 'int32'),
                         ('nsph', 'int32'),
                         ('nstar', 'int32'),
                         ('aa', 'float'),
                         ('gravsmooth', 'float')])
                header = numpy.zeros((), dtype=dtype)
                header['eflag'] = 1
                header['hsize'] = 20
                header['npart'] = size
                header.tofile(ff)

        self.comm.barrier()

        with open(output, 'r+b') as ff:
            ff.seek(28 + offset * 12)
            numpy.float32(subsample['Position']).tofile(ff)
            ff.seek(28 + offset * 12 + size * 12)
            numpy.float32(subsample['Velocity']).tofile(ff)
            ff.seek(28 + offset * 4 + size * 24)
            numpy.float32(subsample['Density']).tofile(ff)
            ff.seek(28 + offset * 8 + size * 28)
            numpy.uint64(subsample['ID']).tofile(ff)



import numpy

from scipy.interpolate import InterpolatedUnivariateSpline
class interp1d(InterpolatedUnivariateSpline):
  """ this replaces the scipy interp1d which do not always
      pass through the points
      note that kind has to be an integer as it is actually
      a UnivariateSpline.
  """
  def __init__(self, x, y, kind, bounds_error=False, fill_value=numpy.nan, copy=True):
    if copy:
      self.x = x.copy()
      self.y = y.copy()
    else:
      self.x = x
      self.y = y
    InterpolatedUnivariateSpline.__init__(self, self.x, self.y, k=kind)
    self.xmin = self.x[0]
    self.xmax = self.x[-1]
    self.fill_value = fill_value
    self.bounds_error = bounds_error
  def __call__(self, x):
    x = numpy.asarray(x)
    shape = x.shape
    x = x.ravel()
    bad = (x > self.xmax) | (x < self.xmin)
    if self.bounds_error and numpy.any(bad):
      raise ValueError("some values are out of bounds")
    y = InterpolatedUnivariateSpline.__call__(self, x.ravel())
    y[bad] = self.fill_value
    return y.reshape(shape)

def corrfrompower(K, P, R, res=100000):
    """calculate correlation function from power spectrum,
       P is 1d powerspectrum. if R is not None, estimate at
       those points .
       returns R, xi(R)
       internally this does a numerical integral with the trapz
       rule for the radial direction of the fourier transformation,
       with a gaussian damping kernel (learned from Xiaoying) 
       the nan points of P is skipped. (usually when K==0, P is nan)
       input power spectrum is assumed to have the gadget convention,
       AKA, normalized to (2 * pi) ** -3 times sigma_8.
       The integral can also be done on log K instead of K if logscale 
       is True.
    """
    mask = ~numpy.isnan(P) & (K > 0)
    K = K[mask]
    P = P[mask]
    Pfunc = interp1d(K, P, kind=5)
    K = numpy.linspace(K.min(), K.max(), int(res))
    P = Pfunc(K)

    # use a log scale
    weight = K #* numpy.exp(-K**2)
    diff = numpy.log(K)
    XI = [4 * numpy.pi / r * \
        numpy.trapz(P * numpy.sin(K * r) * K * weight, diff) for r in R]
    XI = (2 * numpy.pi) ** -3 * numpy.array(XI)
  
    return R, XI



from .pipeline import RunAlgorithm, add_run_fixture
from . import os, asserts, unittest
from .. import examples_dir

class RunBoxSizeAlgorithm(RunAlgorithm):
    run_dir = os.path.join(examples_dir, 'boxsize')
           

@add_run_fixture(__name__, RunBoxSizeAlgorithm, 'TestBoxSize')
class TestBoxSize(unittest.TestCase):
    param_file  = "test_boxsize.params"
    output_file = "test_boxsize.dat"
    datasources = ['test_bianchi_randoms.dat']
    
    def test_exit_code(self):
        asserts.test_exit_code(self)
    
    def test_exception(self):
        asserts.test_exception(self)    
    
    def test_result(self):
        asserts.test_result_md5sum(self)




from nbodykit.extensionpoints import DataSource
import numpy
import logging
import bigfile

logger = logging.getLogger('HaloLabel')

class HaloLabel(DataSource):
    """
    DataSource for reading a file of halo labels (halo id per particle), 
    as generated the FOF algorithm
    """
    plugin_name = "HaloLabel"
    
    def __init__(self, path, bunchsize=4*1024*1024):
        f = bigfile.BigFileMPI(self.comm, self.path)
        self.size = f['Label'].size
    
    @classmethod
    def register(cls):
        
        s = cls.schema
        s.description = "read a file of halo labels (halo id per particle), as generated the FOF algorithm"
        s.add_argument("path", type=str, help="the file path to load the data from")
        s.add_argument("bunchsize", type=int, help="number of particle to read in a bunch")

    def parallel_read(self, columns, full=False):
        f = bigfile.BigFileMPI(self.comm, self.path)
        readcolumns = set(columns)
        
        # remove columns not in the file (None will be returned)
        for col in list(readcolumns):
            if col not in f:
                readcolumns.remove(col)
        
        done = False
        i = 0
        while not numpy.all(self.comm.allgather(done)):
            ret = []
            dataset = bigfile.BigData(f, readcolumns)

            Ntot = dataset.size
            start = self.comm.rank * Ntot // self.comm.size
            end = (self.comm.rank + 1) * Ntot // self.comm.size

            if not full:
                bunchstart = start + i * self.bunchsize
                bunchend = start + (i + 1) * self.bunchsize
                if bunchend > end: bunchend = end
                if bunchstart > end: bunchstart = end
            else:
                bunchstart = start
                bunchend = end

            if bunchend == end:
                done = True

            P = {}

            for column in readcolumns:
                data = dataset[column][bunchstart:bunchend]
                P[column] = data

            i = i + 1
            yield [P.get(column, None) for column in columns]



import sys
import contextlib
import numpy

class MeasurementStorage(object):
    """
    Class to write a 1D or 2D measurement to a plaintext file
    """    
    def __init__(self, path):
        self.path = path

    @classmethod
    def create(cls, dim, path):
        if dim == '1d':
            return Measurement1DStorage(path)
        elif dim == '2d':
            return Measurement2DStorage(path)
        else:
            raise ValueError("storage dimension must be `1d` or `2d`")
        
    @contextlib.contextmanager
    def open(self):
        if self.path and self.path != '-':
            ff = open(self.path, 'wb')
        else:
            ff = sys.stdout
            
        try:
            yield ff
        finally:
            if ff is not sys.stdout:
                ff.close()

    def write(self, edges, cols, data, **meta):
        return NotImplemented
        
class Measurement1DStorage(MeasurementStorage):

    def write(self, edges, cols, data, **meta):
        """
        Write out a 1D measurement to file
        
        Notes
        -----
        Any lines not holding data values i.e., `edges` or `metadata`,
        will start with the `#` character. This allows the output
        file to be read using the default arguments of `numpy.loadtxt`,
        which treats lines beginning with `#` as comments and skips them. 
        
        Parameters
        ----------
        edges : array_like
            the edges of the bins used for the measurement
        cols : list
            list of names for the corresponding `data`
        data : list of arrays
            list of 1D arrays specifying the data, which are written 
            as columns to file; complex arrays will be written as two
            columns (real and imag).
        meta : 
            Any additional metadata to write to file, specified as keyword 
            arguments
        """
        if len(cols) != len(data):
            raise ValueError("size mismatch between column names and data arrays")
            
        with self.open() as ff:
            
            # split any complex fields into separate columns
            columns = []
            names = []
            for name, column in zip(cols, data):
                if numpy.iscomplexobj(column):
                    columns.append(column.real)
                    columns.append(column.imag)
                    names.append(name + '.real')
                    names.append(name + '.imag')
                else:
                    columns.append(column)
                    names.append(name)

            # write out column names first
            ff.write(("# "+" ".join(names) + "\n").encode())
                     
            # write out the 1D data arrays
            columns = numpy.vstack(columns).T
            numpy.savetxt(ff, columns, '%s')

            # write out the bin edges
            header = "# edges %d\n" %len(edges)
            values = "".join("# %0.7g\n" %e for e in edges)
            ff.write((header+values).encode())
            
            # lastly, write out metadata, if any
            if len(meta):
                ff.write(("# metadata %d\n" %len(meta)).encode())
                for k,v in meta.items():
                    if not numpy.isscalar(v):
                        name = self.__class__.__name__
                        raise NotImplementedError("%s cannot write out non-scalar metadata yet" %name)
                    ff.write(("# %s %s %s\n" %(k, str(v), type(v).__name__)).encode())
            ff.flush()
            
            
class Measurement2DStorage(MeasurementStorage):

    def write(self, edges, cols, data, **meta):
        """
        Write a 2D measurement as plain text.
                
        Parameters
        ----------
        edges : list 
            list specifying the bin edges in both dimensions
        cols : list
            list of names for the corresponding `data`
        data : list
            list of 2D arrays holding the data; complex arrays will be
            treated as two arrays (real and imag).
        meta : dict
            any additional metadata to write out at the end of the file
        """
        if len(cols) != len(data):
            raise ValueError("size mismatch between column names and data arrays")
            
        with self.open() as ff:
            
            # write number of mu and k bins first
            N1, N2 = data[0].shape
            ff.write(("%d %d\n" %(N1, N2)).encode())
            
            # split any complex fields into separate columns
            columns = []
            names = []
            for name, column in zip(cols, data):
                if numpy.iscomplexobj(column):
                    columns.append(column.real)
                    columns.append(column.imag)
                    names.append(name + '.real')
                    names.append(name + '.imag')
                else:
                    columns.append(column)
                    names.append(name)

            # write out column names
            ff.write((" ".join(names) + "\n").encode())
            
            # write out flattened columns
            columns = numpy.dstack(columns).reshape(-1, len(columns))
            numpy.savetxt(ff, columns, '%s')
            
            # write out edges                
            for name, e in zip(['kedges', 'muedges'], edges):
                header = "%s %d\n" %(name, len(e))
                values = "".join("%0.7g\n" %x for x in e)
                ff.write((header+values).encode())
            
            # lastly, write out metadata, if any
            if len(meta):
                ff.write(("metadata %d\n" %len(meta)).encode())
                for k,v in meta.items():
                    if not numpy.isscalar(v):
                        name = self.__class__.__name__
                        raise NotImplementedError("%s cannot write out non-scalar metadata yet" %name)
                    ff.write(("%s %s %s\n" %(k, str(v), type(v).__name__)).encode())
            
            ff.flush()


from nbodykit.extensionpoints import DataSource
from nbodykit.utils import selectionlanguage
import logging
import numpy
import itertools

logger = logging.getLogger('RaDecRedshift')

    
class RaDecRedshiftDataSource(DataSource):
    """
    DataSource designed to handle reading (ra, dec, redshift)
    from a plaintext file, using `pandas.read_csv`
    
    *   Returns the Cartesian coordinates corresponding to 
        (ra, dec, redshift) as the `Position` column.
    *   If `unit_sphere = True`, the Cartesian coordinates
        are on the unit sphere, so the the redshift information
        is not used
    """
    plugin_name = "RaDecRedshift"
    
    def __init__(self, path, names, unit_sphere=False, 
                    usecols=None, sky_cols=['ra','dec'], z_col='z', 
                    weight_col=None, degrees=False, select=None, nbar_col=None):       

        # setup the cosmology
        if not self.unit_sphere:
            if self.cosmo is None:
                raise ValueError("please specify a input Cosmology to use in `RaDecRedshift`")
            
            # sample the cosmology's comoving distance
            self.cosmo.sample('comoving_distance', numpy.logspace(-5, 1, 1024))
        else:
            # unit sphere fits in box of size L = 2
            self.BoxSize = numpy.array([2., 2., 2.])

  
    @classmethod
    def register(cls):
        
        s = cls.schema
        s.description = "read (ra, dec, z) from a plaintext file, returning Cartesian coordinates"
        
        # required
        s.add_argument("path", type=str, help="the file path to load the data from")
        s.add_argument("names", type=str, nargs='+', help="the names of columns in text file")
        
        # optional
        s.add_argument('unit_sphere', type=bool, 
            help='if True, return Cartesian coordinates on the unit sphere')
        s.add_argument("usecols", type=str, nargs='*', 
            help="only read these columns from file")
        s.add_argument("sky_cols", type=str, nargs='*',
            help="names of the columns specifying the sky coordinates")
        s.add_argument("z_col", type=str,
            help="name of the column specifying the redshift coordinate")
        s.add_argument("weight_col", type=str,
            help="name of the column specifying the `weight` for each object")
        s.add_argument("nbar_col", type=str,
            help="name of the column specifying the `nbar` value for each object")
        s.add_argument('degrees', type=bool,
            help='set this flag if the input (ra, dec) are in degrees')
        s.add_argument("select", type=selectionlanguage.Query, 
            help='row selection based on conditions specified as string')
                  
    def _to_cartesian(self, coords):
        """
        Convert the (ra, dec, redshift) coordinates to cartesian coordinates,
        scaled to the comoving distance if `unit_sphere = False`, else
        on the unit sphere
        
        Notes
        -----
        Input angles `ra` and `dec` (first 2 columns of `coords`)  
        are assumed to be in radians
        
        Parameters
        -----------
        coords : array_like, (N, 3)
            the input coordinates with the columns giving (ra, dec, redshift),
            where ``ra`` and ``dec`` are in radians
        
        Returns
        -------
        pos : array_like, (N,3)
            the cartesian position coordinates, where columns represent ``x``, 
            ``y``, and ``z``
        """
        ra, dec, redshift = coords.T
        
        x = numpy.cos( dec ) * numpy.cos( ra )
        y = numpy.cos( dec ) * numpy.sin( ra )
        z = numpy.sin( dec )
                
        pos = numpy.vstack([x,y,z])
        if not self.unit_sphere:
            pos *= self.cosmo.comoving_distance(redshift)
        
        return pos.T
        
    def readall(self):
        """
        Read all available data, returning a dictionary
        
        This provides the following columns:
        
            ``Ra``  : right ascension (in radians)
            ``Dec`` : declination (in radians)  
            ``Redshift`` : the redshift
            ``Position`` : cartesian coordinates computed from angular coords + redshift
        
        And optionally, the `Weight` and `Nbar` columns
        """  
        try:
            import pandas as pd
        except:
            name = self.__class__.__name__
            raise ImportError("pandas must be installed to use %s" %name)

        # read in the plain text file using pandas
        kwargs = {}
        kwargs['comment'] = '#'
        kwargs['names'] = self.names
        kwargs['header'] = None
        kwargs['engine'] = 'c'
        kwargs['delim_whitespace'] = True
        kwargs['usecols'] = self.usecols
        
        # iterate through in parallel
        data = pd.read_csv(self.path, **kwargs)

        # select based on input conditions
        if self.select is not None:
            mask = self.select.get_mask(data)
            data = data[mask]

        # rescale the angles
        if self.degrees:
            data[self.sky_cols] *= numpy.pi/180.

        # get the (ra, dec, z) coords
        cols = self.sky_cols + [self.z_col]
        pos = data[cols].values.astype('f4')

        toret             = {}
        toret['Ra']       = pos[:,0]
        toret['Dec']      = pos[:,1]
        toret['Redshift'] = pos[:,2]
        toret['Position'] = self._to_cartesian(pos)

        # optionally, return a weight
        if self.weight_col is not None:
            toret['Weight'] = data[self.weight_col].values.astype('f4')

        # optionally, return nbar
        if self.nbar_col is not None:
            toret['Nbar'] = data[self.nbar_col].values.astype('f4')
    
        return toret


import numpy
from mpi4py import MPI
import logging
from nbodykit.extensionpoints import Painter, Transfer
import time

logger = logging.getLogger('measurestats')

def timer(start, end):
    hours, rem = divmod(end-start, 3600)
    minutes, seconds = divmod(rem, 60)
    return "{:0>2}:{:0>2}:{:05.2f}".format(int(hours),int(minutes),seconds)

def compute_3d_power(fields, pm, comm=None):
    """
    Compute and return the 3D power from two input fields
    
    Parameters
    ----------
    fields : list of (``DataSource``, ``Painter``, ``Transfer``) tuples
        the list of fields which the 3D power will be computed
        
    pm : ``ParticleMesh``
        particle mesh object that handles the painting and FFTs
                
    comm : MPI.Communicator, optional
        the communicator to pass to the ``ParticleMesh`` object. If not
        provided, ``MPI.COMM_WORLD`` is used
        
    Returns
    -------
    p3d : array_like (real)
        the 3D power spectrum, corresponding to the gridded input fields
        
    stats1 : dict
        dictionary holding the statistics of the first field, as returned
        by the `Painter` used 
        
    stats2 : dict
        dictionary holding the statistics of the second field, as returned
        by the `Painter` used
    """
    # some setup
    rank = comm.rank if comm is not None else MPI.COMM_WORLD.rank
    
    # check that the painter was passed correctly
    datasources = [d for d, p, t in fields]
    painters    = [p for d, p, t in fields]
    transfers   = [t for d, p, t in fields]
    
    # paint, FT field and filter field #1
    stats1 = painters[0].paint(pm, datasources[0])
    if rank == 0: logger.info('%s painting done' %pm.paintbrush)
    pm.r2c()
    if rank == 0: logger.info('r2c done')
    pm.transfer(transfers[0])

    # do the cross power if two fields supplied
    if len(fields) > 1:
                
        # crash if box size isn't the same
        if not numpy.all(datasources[0].BoxSize == datasources[1].BoxSize):
            raise ValueError("mismatch in box sizes for cross power measurement")
        
        # copy and store field #1's complex
        c1 = pm.complex.copy()
        
        # paint, FT, and filter field #2
        stats2 = painters[1].paint(pm, datasources[1])
        if rank == 0: logger.info('%s painting 2 done' %pm.paintbrush)
        pm.r2c()
        if rank == 0: logger.info('r2c 2 done')
        pm.transfer(transfers[1])
        c2 = pm.complex
  
    # do the auto power
    else:
        c1 = pm.complex
        c2 = pm.complex
        stats2 = stats1

    # reuse the memory in c1.real for the 3d power spectrum
    p3d = c1
    
    # calculate the 3d power spectrum, islab by islab to save memory
    for islab in range(len(c1)):
        p3d[islab, ...] = c1[islab]*c2[islab].conj()

    # the complex field is dimensionless; power is L^3
    # ref to http://icc.dur.ac.uk/~tt/Lectures/UA/L4/cosmology.pdf
    p3d[...] *= pm.BoxSize.prod() 
                
    return p3d, stats1, stats2
    
def compute_bianchi_poles(comm, max_ell, catalog, Nmesh, 
                            factor_hexadecapole=False, 
                            paintbrush='cic'):
    """
    Compute and return the 3D power multipoles (ell = [0, 2, 4]) from one 
    input field, which contains non-trivial survey geometry.
    
    The estimator uses the FFT algorithm outlined in Bianchi et al. 2015 to compute
    the monopole, quadrupole, and hexadecapole
    
    Parameters
    ----------
    max_ell : int
        the maximum multipole number to compute up to (and including). Must be
        one of [0, 2, 4]
        
    datasource : ``DataSource``
        the DataSource which will be return the Cartesian coordinates of the
        (ra, dec, z) survey
        
    pm : ``ParticleMesh``
        particle mesh object that handles the painting and FFTs
            
    factor_hexadecapole: bool, optional
        if `True`, use the factored expression for the hexadecapole (ell=4) from
        eq. 27 of Scoccimarro 2015 (1506.02729); default is `False`
    
    Returns
    -------
    poles : list
        list of the 3D power multipoles computed, up to and including `ell=max_ell`
        
    stats : dict
        dictionary holding the statistics of the input datasource, as returned
        by the `FKPPainter` painter
    """
    from pmesh.particlemesh import ParticleMesh
    
    def bianchi_transfer(data, x, i, j, k=None, offset=[0., 0., 0.]):
        """
        Transfer functions necessary to compute the power spectrum  
        multipoles via FFTs
        
        This multiplies by one of two kernels:
        
            1. x_i * x_j / x**2 * data, if `k` is None
            2. x_i**2 * x_j * x_k / x**4 * data, if `k` is not None
        
        See equations 10 (for quadrupole) and 12 (for hexadecapole)
        of Bianchi et al 2015.
        
        Parameters
        ----------
        data : array_like
            the array to rescale -- either the configuration-space 
            `pm.real` or the Fourier-space `pm.complex`
        x : array_like
            the coordinate array -- either `pm.r` or `pm.k`
        i, j, k : int
            the integers specifying the coordinate axes; see the 
            above description 
        offset : array_like
            add an average offset to the `x` component arrays; 
            needed if the simulation box has an average offset 
            from the grid box in configuration-space
        """
        # compute x**2
        norm = sum((xi + offset[ii])**2 for ii, xi in enumerate(x))
    
        # get x_i, x_j
        # if i == 'x' direction, it's just one value
        xi = x[i] + offset[i]
        xj = x[j] + offset[j]

        # multiply the kernel
        if k is not None:
            xk = x[k] + offset[k]
            data[:] *= xi**2 * xj * xk
            idx = norm != 0.
            data[idx] /= norm[idx]**2
        
        else:
            data[:] *= xi * xj
            idx = norm != 0.
            data[idx] /= norm[idx]
    
    # the rank
    rank = comm.rank
    
    # the painting kernel transfer
    if paintbrush == 'cic':
        transfer = Transfer.create('AnisotropicCIC')
    elif paintbrush == 'tsc':
        transfer = Transfer.create('AnisotropicTSC')
    else:
        raise ValueError("valid `paintbrush` values are: ['cic', 'tsc']")

    # which transfers do we need
    if max_ell not in [0, 2, 4]:
        raise ValueError("valid values for the maximum multipole number are [0, 2, 4]")
    ells = numpy.arange(0, max_ell+1, 2)
    bianchi_transfers = []
    
    # quadrupole kernels
    if max_ell > 0:
        k2 = [(0, 0), (1, 1), (2, 2), (0, 1), (0, 2), (1, 2)]
        a2 = [1.]*3 + [2.]*3
        bianchi_transfers.append((a2, k2))
    
    # hexadecapole kernels
    if max_ell > 2 and not factor_hexadecapole:
        k4 = [(0, 0, 0), (1, 1, 1), (2, 2, 2), (0, 0, 1), (0, 0, 2),
             (1, 1, 0), (1, 1, 2), (2, 2, 0), (2, 2, 1), (0, 1, 1),
             (0, 2, 2), (1, 2, 2), (0, 1, 2), (1, 0, 2), (2, 0, 1)]
        a4 = [1.]*3 + [4.]*6 + [6.]*3 + [12.]*3
        bianchi_transfers.append((a4, k4))
    
    # load the data/randoms and setup boxsize, etc
    with catalog:
        
        # the mean coordinate offset
        offset = catalog.mean_coordinate_offset
        
        # initialize the particle mesh
        pm = ParticleMesh(catalog.BoxSize, Nmesh, paintbrush=paintbrush, dtype='f8', comm=comm)
        
        # do the FKP painting
        stats = catalog.paint(pm)
    
    # save the fkp density for later
    density = pm.real.copy()
    if rank == 0: logger.info('%s painting done' %paintbrush)
    
    # compute the monopole, A0(k), and save
    pm.r2c()
    transfer(pm, pm.complex)
    A0 = pm.complex.copy()
    if rank == 0: logger.info('ell = 0 done; 1 r2c completed')
    
    volume = pm.BoxSize.prod()
    
    # store the A0, A2, A4 arrays
    result = []
    result.append(A0*volume)
    
    # the x grid points (at point centers)
    cell_size = pm.BoxSize / pm.Nmesh
    xgrid = [(ri+0.5)*cell_size[i] for i, ri in enumerate(pm.r)]
    
    start = time.time()
    for iell in range(len(bianchi_transfers)):
        ell = ells[iell+1]
        A_ell = numpy.zeros_like(pm.complex)
        
        for amp, integers in zip(*bianchi_transfers[iell]):
                        
            # reset the 'pm.real' array to the original FKP density
            pm.real[:] = density[:]
        
            # apply the real-space transfer
            if rank == 0: logger.debug("applying real-space Bianchi transfer for %s..." %str(integers))
            bianchi_transfer(pm.real, xgrid, *integers, offset=offset)
            if rank == 0: logger.debug('...done')
    
            # do the FT and apply the k-space kernel
            if rank == 0: logger.debug("performing r2c...")
            pm.r2c()
            if rank == 0: logger.debug('...done')
            
            # fourier space kernel
            if rank == 0: logger.debug("applying Fourier-space Bianchi transfer for %s..." %str(integers))
            bianchi_transfer(pm.complex, pm.k, *integers)
            if rank == 0: logger.debug('...done')
            
            # and save
            A_ell[:] += amp*pm.complex[:]*volume
            
        # apply the gridding transfer and save
        transfer(pm, A_ell)
        result.append(A_ell)
        if rank == 0: 
            args = (ell, len(bianchi_transfers[iell][0]))
            logger.info('ell = %d done; %s r2c completed' %args)
        
    stop = time.time()
    if rank == 0:
        logger.info("higher order multipoles computed in elapsed time %s" %timer(start, stop))
        if factor_hexadecapole:
            logger.info("using factorized hexadecapole estimator for ell=4")
    
    # proper normalization: A_ran from equation 
    norm = 1.0 / stats['A_ran']
    
    # reuse memory for output
    P0 = result[0]
    if max_ell > 0: P2 = result[1]
    if max_ell > 2:
        if not factor_hexadecapole: 
            P4 = result[2]
        else:
            P4 = numpy.empty_like(P2)
        
    # calculate the multipoles, islab by islab to save memory
    # see equations 6-8 of Bianchi et al. 2015
    for islab in range(len(P0)):

        # save for reuse
        P0_star = (P0[islab]).conj()
        if max_ell > 0: P2_star = (P2[islab]).conj()

        # hexadecapole
        if max_ell > 2:
            if not factor_hexadecapole:
                P4[islab, ...] = norm * 9./8. * P0[islab] * (35.*(P4[islab]).conj() - 30.*P2_star + 3.*P0_star)
            else:
                P4[islab, ...] = norm * 9./8. * ( 35.*P2[islab]*P2_star + 3.*P0[islab]*P0_star - 5./3.*(11.*P0[islab]*P2_star + 7.*P2[islab]*P0_star) )
        # quadrupole
        if max_ell > 0:
            P2[islab, ...] = norm * 5./2. * P0[islab] * (3.*P2_star - P0_star)

        # monopole
        P0[islab, ...] = norm * P0[islab] * P0_star
        
    result = [P0]
    if max_ell > 0: result.append(P2)
    if max_ell > 2: result.append(P4)

    return pm, result, stats


def compute_brutal_corr(datasources, rbins, Nmu=0, comm=None, subsample=1, los='z', poles=[]):
    """
    Compute the correlation function by direct pair summation, projected
    into either 1d `R` bins or 2d (`R`, `mu`) bins
    
    Parameters
    ----------
    datasources : list of ``DataSource`` objects
        the list of datasources from which the 3D correlation will be computed
        
    Rmax : float
        the maximum R value to compute, in the same units as the input
        datasources
    
    Nmu : int, optional
        the number of desired `mu` bins, where `mu` is the cosine 
        of the angle from the line-of-sight. Default is 0, in 
        which case the correlation function is binned as a function of 
        `R` only
        
    comm : MPI.Communicator, optional
        the communicator to pass to the ``ParticleMesh`` object. If not
        provided, ``MPI.COMM_WORLD`` is used
        
    subsample : int, optional
        Down-sample the input datasources by choosing 1 out of every N points. 
        Default is `1`
    
    los : 'x', 'y', 'z', optional
        the dimension to treat as the line-of-sight; default is 'z'
        
    Returns
    -------
    pc : ``correlate.paircount``
        the ``kdcount`` pair counting instance
    xi : array_like
        the correlation function
    RR : array_like
        the number of random-random pairs (used as normalization of the data-data pairs)
    """
    from pmesh.domain import GridND
    from kdcount import correlate
    
    if los not in "xyz":
        raise ValueError("the `los` must be one of `x`, `y`, or `z`")
    los = "xyz".index(los)
    poles = numpy.array(poles)
    Rmax = rbins[-1]
    
    # the comm
    if comm is None: comm = MPI.COMM_WORLD
    
    # determine processors for grididng
    for Nx in range(int(comm.size**0.3333) + 1, 0, -1):
        if comm.size % Nx == 0: break
    else:
        Nx = 1
    for Ny in range(int(comm.size**0.5) + 1, 0, -1):
        if (comm.size // Nx) % Ny == 0: break
    else:
        Ny = 1
    Nz = comm.size // Nx // Ny
    Nproc = [Nx, Ny, Nz]
    if comm.rank == 0:
        logger.info('Nproc = %s' %str(Nproc))
        logger.info('Rmax = %g' %Rmax)
    
    # domain decomposition
    grid = [numpy.linspace(0, datasources[0].BoxSize[i], Nproc[i]+1, endpoint=True) for i in range(3)]
    domain = GridND(grid, comm=comm)

    # read position for field #1 
    with datasources[0].open() as stream:
        [[pos1]] = stream.read(['Position'], full=True)
    pos1 = pos1[comm.rank * subsample // comm.size ::subsample]
    N1 = comm.allreduce(len(pos1))
    
    # read position for field #2
    if len(datasources) > 1:
        with datasources[1].open() as stream:
            [[pos2]] = stream.read(['Position'], full=True)
        pos2 = pos2[comm.rank * subsample // comm.size ::subsample]
        N2 = comm.allreduce(len(pos2))
    else:
        pos2 = pos1
        N2 = N1
    
    # exchange field #1 positions    
    layout = domain.decompose(pos1, smoothing=0)
    pos1 = layout.exchange(pos1)
    if comm.rank == 0:
        logger.info('exchange pos1')
        
    # exchange field #2 positions
    if Rmax > datasources[0].BoxSize[0] * 0.25:
        pos2 = numpy.concatenate(comm.allgather(pos2), axis=0)
    else:
        layout = domain.decompose(pos2, smoothing=Rmax)
        pos2 = layout.exchange(pos2)
    if comm.rank == 0:
        logger.info('exchange pos2')

    # initialize the points trees
    tree1 = correlate.points(pos1, boxsize=datasources[0].BoxSize)
    tree2 = correlate.points(pos2, boxsize=datasources[0].BoxSize)

    logger.info('rank %d correlating %d x %d' %(comm.rank, len(tree1), len(tree2)))
    if comm.rank == 0:
        logger.info('all correlating %d x %d' %(N1, N2))

    # the binning, either r or (r,mu)
    if len(poles):
        bins = correlate.FlatSkyMultipoleBinning(rbins, poles, los, compute_mean_coords=True)
    elif Nmu > 0:
        bins = correlate.FlatSkyBinning(rbins, Nmu, los, compute_mean_coords=True)
    else:
        bins = correlate.RBinning(rbins, compute_mean_coords=True)

    # do the pair count
    # have to set usefast = False to get mean centers, or exception thrown
    pc = correlate.paircount(tree2, tree1, bins, np=0, usefast=False)
    pc.sum1[:] = comm.allreduce(pc.sum1)
    
    # get the mean bin values, reducing from all ranks
    pc.pair_counts[:] = comm.allreduce(pc.pair_counts)
    with numpy.errstate(invalid='ignore'):
        if bins.Ndim > 1:
            for i in range(bins.Ndim):
                pc.mean_centers[i][:] = comm.allreduce(pc.mean_centers_sum[i]) / pc.pair_counts
        else:
            pc.mean_centers[:] = comm.allreduce(pc.mean_centers_sum[0]) / pc.pair_counts

    # compute the random pairs from the fractional volume
    RR = 1.*N1*N2 / datasources[0].BoxSize.prod()
    if Nmu > 0:
        dr3 = numpy.diff(pc.edges[0]**3)
        dmu = numpy.diff(pc.edges[1])
        RR *= 2. / 3. * numpy.pi * dr3[:,None] * dmu[None,:]
    else:
        RR *= 4. / 3. * numpy.pi * numpy.diff(pc.edges**3)
    
    # return the correlation and the pair count object
    xi = (1. * pc.sum1 / RR) - 1.0
    if len(poles):
        xi = xi.T # make ell the second axis 
        xi[:,poles!=0] += 1.0 # only monopole gets minus one

    return pc, xi, RR

def compute_3d_corr(fields, pm, comm=None):
    """
    Compute the 3d correlation function by Fourier transforming 
    the 3d power spectrum
    
    See documentation of `measurestats.compute_3d_power`
    """

    p3d, N1, N2 = compute_3d_power(fields, pm, comm=comm)
    
    # directly transform dimensionless p3d
    # Note that L^3 cancels with dk^3.
    pm.complex[:] = p3d.copy()
    pm.complex[:] *= 1.0 / pm.BoxSize.prod()
    pm.c2r()
    xi3d = pm.real
    
    return xi3d, N1, N2


def project_to_basis(comm, x3d, y3d, edges, los='z', poles=[], symmetric=True):
    """ 
    Project a 3D statistic on to the specified basis. 

    The projection will be onto 2d bins `(x, mu)`, where `x`
    is separation `r` in configuration space or wavenumber `k` in 
    Fourier space, and `mu` is the cosine of the angle to the 
    line-of-sight. 
    
    Optionally, the multipoles of the 2d `(x, mu)` bins are 
    also returned, as specified by the multipole numbers in `poles`
    
    Notes
    -----
    *   the mu range extends from 0 to 1.0
    *   the mu bins are half-inclusive half-exclusive, except the last bin
        is inclusive on both ends (to include mu = 1.0)
    *   when Nmu == 1, the case reduces to the isotropic 1D binning
    
    Parameters
    ----------
    comm : MPI.Comm
        the communicator for the decomposition of the power spectrum
        
    x3d  : list
        The list of x values for each item in the `y3d` array. The items
        must broadcast to the same shape of `y3d`.

    y3d : array_like (real)
        a 3D statistic, either a power spectrum (defined in Fourier space),
        or a correlation function (defined in configuration space)

    edges : array_like
        an array specifying the edges of the `x` bins, where `x` is 
        either Fourier space `k` or configuration space `r`

    Nmu : int
        the number of mu bins to use when binning in the 3d statistic
    
    los : str, {'x','y','z'}
        the line-of-sight direction, which the angle `mu` is defined with
        respect to. Default is `z`.
        
    poles : list of int, optional
        if provided, a list of integers specifying multipole numbers to
        project the 2d `(x, mu)` on to
        
    symmetric : bool, optional
        If `True`, the `y3d` area is assumed to be symmetric about the `z = 0`
        plane. If `y3d` is a power spectrum, this should be set to `True`, 
        while if `y3d` is a correlation function, this should be `False`        
    """
    from scipy.special import legendre

    # bin edges
    xedges, muedges = edges
    Nx = len(xedges) - 1 
    Nmu = len(muedges) - 1
    
    # always measure make sure first ell is monopole, which
    # is just (x, mu) projection since legendre of ell=0 is 1
    do_poles = len(poles) > 0
    poles_ = [0]+sorted(poles) if 0 not in poles else sorted(poles)
    legpoly = [legendre(l) for l in poles_]
    ell_idx = [poles_.index(l) for l in poles]
    Nell = len(poles_)
    
    # valid ell values
    if any(ell < 0 for ell in poles_):
        raise RuntimeError("multipole numbers must be nonnegative integers")

    # squared x bin edges
    xedges2 = xedges ** 2

    musum = numpy.zeros((Nx+2, Nmu+2))
    xsum = numpy.zeros((Nx+2, Nmu+2))
    ysum = numpy.zeros((Nell, Nx+2, Nmu+2), dtype=y3d.dtype) # extra dimension for multipoles
    Nsum = numpy.zeros((Nx+2, Nmu+2))
    
    # los index
    los_index = 'xyz'.index(los)
    
    # need to count all modes with positive z frequency twice due to r2c FFTs
    nonsingular = numpy.squeeze(x3d[2] > 0.) # has length of Nz now

    for islab in range(len(x3d[0])):
        
        # now xslab stores x3d ** 2
        xslab = numpy.float64(x3d[0][islab] ** 2)
        for xi in x3d[1:]:
            xslab = xslab + xi[0] ** 2

        if len(xslab.flat) == 0:
            # no data
            continue

        dig_x = numpy.digitize(xslab.flat, xedges2)
    
        # make xslab just x
        xslab **= 0.5
    
        # store mu (keeping track of positive/negative)
        with numpy.errstate(invalid='ignore'):
            if los_index == 0:
                mu = x3d[los_index][islab]/xslab
            else:
                mu = x3d[los_index][0]/xslab
        dig_mu = numpy.digitize(abs(mu).flat, muedges)
        
        # make the multi-index
        multi_index = numpy.ravel_multi_index([dig_x, dig_mu], (Nx+2,Nmu+2))
    
        # count modes not in singular plane twice
        if symmetric: xslab[:, nonsingular] *= 2.
    
        # the x sum
        xsum.flat += numpy.bincount(multi_index, weights=xslab.flat, minlength=xsum.size)
    
        # count number of modes
        Nslab = numpy.ones_like(xslab)
        if symmetric: Nslab[:, nonsingular] = 2. # count modes not in singular plane twice
        Nsum.flat += numpy.bincount(multi_index, weights=Nslab.flat, minlength=Nsum.size)

        # weight P(k,mu) and sum for the poles
        for iell, ell in enumerate(poles_):
            
            weighted_y3d = legpoly[iell](mu) * y3d[islab]

            # add conjugate for this kx, ky, kz, corresponding to 
            # the (-kx, -ky, -kz) --> need to make mu negative for conjugate
            # Below is identical to the sum of
            # Leg(ell)(+mu) * y3d[:, nonsingular]    (kx, ky, kz)
            # Leg(ell)(-mu) * y3d[:, nonsingular].conj()  (-kx, -ky, -kz)
            # or 
            # weighted_y3d[:, nonsingular] += (-1)**ell * weighted_y3d[:, nonsingular].conj()
            # but numerically more accurate.
            if symmetric:
                if ell % 2: # odd, real part cancels
                    weighted_y3d.real[:, nonsingular] = 0.
                    weighted_y3d.imag[:, nonsingular] *= 2.
                else:  # even, imag part cancels
                    weighted_y3d.real[:, nonsingular] *= 2.
                    weighted_y3d.imag[:, nonsingular] = 0.
                    
            weighted_y3d *= (2.*ell + 1.)
            ysum[iell,...].real.flat += numpy.bincount(multi_index, weights=weighted_y3d.real.flat, minlength=Nsum.size)
            if numpy.iscomplexobj(ysum):
                ysum[iell,...].imag.flat += numpy.bincount(multi_index, weights=weighted_y3d.imag.flat, minlength=Nsum.size)
        
        # the mu sum
        if symmetric: mu[:, nonsingular] *= 2.
        musum.flat += numpy.bincount(multi_index, weights=abs(mu).flat, minlength=musum.size)

    xsum  = comm.allreduce(xsum, MPI.SUM)
    musum = comm.allreduce(musum, MPI.SUM)
    ysum  = comm.allreduce(ysum, MPI.SUM)
    Nsum  = comm.allreduce(Nsum, MPI.SUM)

    # add the last 'internal' mu bin (mu == 1) to the last visible mu bin
    # this makes the last visible mu bin inclusive on both ends.
    ysum[..., -2] += ysum[..., -1]
    musum[:, -2]  += musum[:, -1]
    xsum[:, -2]   += xsum[:, -1]
    Nsum[:, -2]   += Nsum[:, -1]

    # reshape and slice to remove out of bounds points
    sl = slice(1, -1)
    with numpy.errstate(invalid='ignore'):
        
        # projected results
        y2d       = (ysum[0,...] / Nsum)[sl,sl] # ell=0 is first index
        xmean_2d  = (xsum / Nsum)[sl,sl]
        mumean_2d = (musum / Nsum)[sl, sl]
        N_2d      = Nsum[sl,sl]
        
        # 1D multipole results (summing over mu (last) axis)
        if do_poles:
            N_1d     = Nsum[sl,sl].sum(axis=-1)
            xmean_1d = xsum[sl,sl].sum(axis=-1) / N_1d
            poles    = ysum[:, sl,sl].sum(axis=-1) / N_1d
            poles    = poles[ell_idx,...]
    
    # return just y(x,mu) or y(x,mu) + multipoles
    result = (xmean_2d, mumean_2d, y2d, N_2d)
    pole_result = (xmean_1d, poles, N_1d) if do_poles else None
    
    return result, pole_result



from nbodykit.extensionpoints import Painter
import numpy
import logging

logger = logging.getLogger('GridPainter')

class GridPainter(Painter):
    """
    Painter that reads a flattened 1D quantity from a binary file, 
    which corresponds to gridded field values on a (Nmesh, Nmesh, Nmesh)
    mesh
    """
    plugin_name = "GridPainter"
    
    def __init__(self):
        pass

    @classmethod
    def register(cls):
        pass
        
    def paint(self, pm, datasource):
        """
        Read the datasource (which specifies the binary file) and
        set ``pm.real`` appropriately given the gridded data
    
        Parameters
        ----------
        pm : ``ParticleMesh``
            particle mesh object that does the painting
        datasource : ``DataSource``
            the data source object representing the field to paint onto the mesh
            
        Returns
        -------
        Ntot : int
            returns ``0`` for the moment, since we don't have info
            on the total number of particles
        """            
        pm.real[:] = 0
        shape = (pm.Nmesh,)*pm.partition.Ndim
        try:
            dtype = numpy.dtype(datasource.dtype)
        except:
            raise ValueError("requested data type `%s` not understood" %datasource.dtype)

        # this gives tuples of integer indices (start, stop) for each dimension
        # that this rank is responsible for 
        slices = [(start, start+width) for start,width in zip(pm.partition.local_i_start, pm.partition.local_ni)]
        
        with open(datasource.path, 'rb') as ff:
        
            for i, a in enumerate(range(*slices[0])):
                for j, b in enumerate(range(*slices[1])):
                    start = numpy.ravel_multi_index([a, b, slices[2][0]], shape)
                    stop = numpy.ravel_multi_index([a, b, slices[2][1]-1], shape)
                    
                    # crash on error - likely Nmesh problem
                    try:
                        ff.seek(start*dtype.itemsize, 0)
                        pm.real[i,j,:] = numpy.fromfile(ff, count=stop-start+1, dtype=dtype)
                    except Exception as e:
                        args = (pm.Nmesh, str(e))
                        raise ValueError("cannot read binary file using `Nmesh = %d`; original message: %s" %args)
                        
        return {}
            


from .pipeline import RunAlgorithm, add_run_fixture
from . import os, asserts, unittest
from .. import examples_dir

class RunZHistAlgorithm(RunAlgorithm):
    run_dir = os.path.join(examples_dir, 'zhist')
           

@add_run_fixture(__name__, RunZHistAlgorithm, 'RedshiftHistogram')
class Test1(unittest.TestCase):
    param_file  = "test_zhist_1.params"
    output_file = "test_zhist_1.dat"
    datasources = ['test_bianchi_randoms.dat']
    
    def test_exit_code(self):
        asserts.test_exit_code(self)
    
    def test_exception(self):
        asserts.test_exception(self)  
        
    def test_result(self):
        asserts.test_result_md5sum(self)
        
        
@add_run_fixture(__name__, RunZHistAlgorithm, 'RedshiftHistogram')
class Test2(unittest.TestCase):
    param_file  = "test_zhist_2.params"
    output_file = "test_zhist_2.dat"
    datasources = ['test_bianchi_randoms.dat']
    
    def test_exit_code(self):
        asserts.test_exit_code(self)
    
    def test_exception(self):
        asserts.test_exception(self)   
        
    def test_result(self):
        asserts.test_result_md5sum(self) 




import numpy
from mpi4py import MPI

from .ndarray import equiv_class

def centerofmass(label, pos, boxsize=1.0, comm=MPI.COMM_WORLD):
    """
    Calulate the center of mass of particles of the same label.

    The center of mass is defined as the mean of positions of particles,
    but care has to be taken regarding to the periodic boundary.

    This is a collective operation, and after the call, all ranks
    will have the position of halos.

    Parameters
    ----------
    label : array_like (integers)
        Halo label of particles, >=0
    pos   : array_like (float, 3)
        position of particles.
    boxsize : float or None
        size of the periodic box, or None if no periodic boundary is assumed.
    comm : :py:class:`MPI.Comm`
        communicator for the collective operation.
    
    Returns
    -------
    hpos : array_like (float, 3)
        the center of mass position of the halos.

    """
    Nhalo0 = max(comm.allgather(label.max())) + 1

    N = numpy.bincount(label, minlength=Nhalo0)
    comm.Allreduce(MPI.IN_PLACE, N, op=MPI.SUM)

    if boxsize is not None:
        posmin = equiv_class(label, pos, op=numpy.fmin, dense_labels=True, identity=numpy.inf,
                        minlength=len(N))
        comm.Allreduce(MPI.IN_PLACE, posmin, op=MPI.MIN)
        dpos = pos - posmin[label]
        bhalf = boxsize * 0.5
        dpos[dpos < -bhalf] += boxsize
        dpos[dpos >= bhalf] -= boxsize
    else:
        dpos = pos
    dpos = equiv_class(label, dpos, op=numpy.add, dense_labels=True, minlength=len(N))
    
    comm.Allreduce(MPI.IN_PLACE, dpos, op=MPI.SUM)
    dpos /= N[:, None]

    if boxsize:
        hpos = posmin + dpos
        hpos %= boxsize
    else:
        hpos = dpos
    return hpos
    
def count(label, comm=MPI.COMM_WORLD):
    """
    Count the number of particles of the same label.

    This is a collective operation, and after the call, all ranks
    will have the particle count.

    Parameters
    ----------
    label : array_like (integers)
        Halo label of particles, >=0
    comm : :py:class:`MPI.Comm`
        communicator for the collective operation.
    
    Returns
    -------
    count : array_like
        the count of number of particles in each halo

    """
    Nhalo0 = max(comm.allgather(label.max())) + 1

    N = numpy.bincount(label, minlength=Nhalo0)
    comm.Allreduce(MPI.IN_PLACE, N, op=MPI.SUM)

    return N
